import warnings

import os
import gc
import datetime
import traceback
import ujson as json
import yaml
from pathlib import Path

from ultralytics import YOLO
import ultralytics.data.build as detection_build
from ultralytics.data.dataset import YOLODataset
import ultralytics.models.yolo.classify.train as train_build
from ultralytics.data.dataset import ClassificationDataset

from PyQt5.QtCore import Qt, QThread, pyqtSignal
from PyQt5.QtWidgets import (QFileDialog, QScrollArea, QMessageBox, QWidget, QVBoxLayout,
                             QLabel, QLineEdit, QDialog, QHBoxLayout, QPushButton, QComboBox,
                             QFormLayout, QTabWidget, QDoubleSpinBox, QGroupBox, QFrame, QSpinBox)

from torch.cuda import empty_cache

from coralnet_toolbox.MachineLearning.Community.cfg import get_available_configs
from coralnet_toolbox.MachineLearning.WeightedDataset import WeightedInstanceDataset
from coralnet_toolbox.MachineLearning.WeightedDataset import WeightedClassificationDataset
from coralnet_toolbox.MachineLearning.EvaluateModel.QtBase import EvaluateModelWorker

from coralnet_toolbox.Icons import get_icon

warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=UserWarning)


# ----------------------------------------------------------------------------------------------------------------------
# Classes
# ----------------------------------------------------------------------------------------------------------------------


class TrainModelWorker(QThread):
    """
    Worker thread for training a model.

    Signals:
        training_started: Emitted when the training starts.
        training_completed: Emitted when the training completes.
        training_error: Emitted when an error occurs during training.
    """
    training_started = pyqtSignal()
    training_completed = pyqtSignal()
    training_error = pyqtSignal(str)

    def __init__(self, params, device):
        """
        Initialize the TrainModelWorker.

        Args:
            params: A dictionary of parameters for training.
            device: The device to use for training (e.g., 'cpu' or 'cuda').
        """
        super().__init__()
        self.params = params
        self.device = device
        self.model = None
        self.model_path = None
        self.weighted = False

    def pre_run(self):
        """
        Set up the model and prepare parameters for training.
        """
        try:
            # Extract model path
            self.model_path = self.params.pop('model', None)
            # Get the weighted flag
            self.weighted = self.params.pop('weighted', False)

            # Determine if ultralytics or community
            if self.model_path in get_available_configs(task=self.params['task']):
                self.model_path = get_available_configs(task=self.params['task'])[self.model_path]
                # Cannot use weighted sampling with community models
                self.weighted = False

            # Use the custom dataset class for weighted sampling
            if self.weighted and self.params['task'] == 'classify':
                train_build.ClassificationDataset = WeightedClassificationDataset
            elif self.weighted and self.params['task'] in ['detect', 'segment']:
                detection_build.YOLODataset = WeightedInstanceDataset

            # Load the model (8.3.141) YOLO handles RTDETR
            self.model = YOLO(self.model_path)

            # Set the task in the model itself
            self.model.task = self.params['task']

            # Freeze layers, freeze encoder
            freeze_layers = self.params.pop('freeze_layers', None)

            if freeze_layers:
                # Calculate the number of layers to freeze
                num_layers = len(self.model.model.model[0:-1])
                num_layers = int(num_layers * freeze_layers)
                freeze_layers = [_ for _ in range(0, num_layers)]
                print(f"Encoder layers frozen ({len(freeze_layers)})")
            else:
                freeze_layers = []

            # Set the freeze parameter for ultralytics
            self.params['freeze'] = freeze_layers

        except Exception as e:
            print(f"Error during setup: {e}\n\nTraceback:\n{traceback.format_exc()}")
            self.training_error.emit(f"Error during setup: {e} (see console log)")
            raise

    def run(self):
        """
        Run the training process in a separate thread.
        """
        try:
            # Emit signal to indicate training has started
            self.training_started.emit()

            # Set up the model and parameters
            self.pre_run()

            # Train the model
            self.model.train(**self.params, device=self.device)

            # Post-run cleanup
            self.post_run()

            # Evaluate the model after training
            self.evaluate_model()

            # Emit signal to indicate training has completed
            self.training_completed.emit()

        except Exception as e:
            print(f"Error during training: {e}\n\nTraceback:\n{traceback.format_exc()}")
            self.training_error.emit(f"Error during training: {e} (see console log)")
        finally:
            self._cleanup()

    def post_run(self):
        """
        Clean up resources after training.
        """
        # Revert to the original dataset class without weighted sampling
        if self.weighted and self.params['task'] == 'classify':
            train_build.ClassificationDataset = ClassificationDataset
        elif self.weighted and self.params['task'] in ['detect', 'segment']:
            detection_build.YOLODataset = YOLODataset

    def evaluate_model(self):
        """
        Evaluate the model after training.
        """
        try:
            # Do not evaluate if the user specifies 
            if not self.params.get('Validation', False):
                return
            
            # Check that there is a test folder
            test_folder = f"{self.params['data']}/test"
            print(f"Note: Looking for test folder: {test_folder}")
            if not os.path.exists(test_folder):
                print("Warning: No test folder found in that location. Skipping evaluation.")
                return
            
            # Create an instance of EvaluateModelWorker and start it
            eval_params = {
                'data': self.params['data'],
                'imgsz': self.params['imgsz'],
                'split': 'test',  # Evaluate on the test set only
                'save_dir': Path(self.params['project']) / self.params['name'] / 'test'
            }

            # Create and start the worker thread
            eval_worker = EvaluateModelWorker(model=self.model,
                                              params=eval_params)

            eval_worker.evaluation_error.connect(self.on_evaluation_error)
            eval_worker.run()  # Run the evaluation synchronously (same thread)
            
        except Exception as e:
            print(f"Error during evaluation: {e}\n\nTraceback:\n{traceback.format_exc()}")
            self.training_error.emit(f"Error during evaluation: {e} (see console log)")

    def on_evaluation_started(self):
        """
        Handle the event when the evaluation starts.
        """
        pass

    def on_evaluation_completed(self):
        """
        Handle the event when the evaluation completes.
        """
        pass

    def on_evaluation_error(self, error_message):
        """
        Handle the event when an error occurs during evaluation.

        Args:
            error_message (str): The error message.
        """
        self.training_error.emit(error_message)

    def _cleanup(self):
        """
        Clean up resources after training.
        """
        del self.model
        gc.collect()
        empty_cache()


class Base(QDialog):
    """
    Dialog for training machine learning models for image classification, object detection,
    and instance segmentation.

    :param main_window: MainWindow object
    :param parent: Parent widget
    """
    def __init__(self, main_window, parent=None):
        super().__init__(parent)
        self.main_window = main_window

        self.setWindowIcon(get_icon("coral.png"))
        self.setWindowTitle("Train Model")
        self.resize(600, 750)  

        # Set window settings
        self.setWindowFlags(Qt.Window |
                            Qt.WindowCloseButtonHint |
                            Qt.WindowMinimizeButtonHint |
                            Qt.WindowMaximizeButtonHint |
                            Qt.WindowTitleHint)

        # Task
        self.task = None
        # For holding parameters
        self.params = {}
        self.custom_params = []
        # Best model weights
        self.model_path = None
        # Class mapping
        self.class_mapping = {}

        # Task specific parameters
        self.imgsz = 640
        self.batch = 4

        # Create the layout
        self.layout = QVBoxLayout(self)

        # Create the info layout
        self.setup_info_layout()
        # Create the dataset layout
        self.setup_dataset_layout()
        # Create the model layout (new)
        self.setup_model_layout()
        # Create the output layout
        self.setup_output_layout()
        # Create and set up the parameters layout
        self.setup_parameters_layout()
        # Create the buttons layout
        self.setup_buttons_layout()

    def setup_info_layout(self):
        """
        Set up the layout and widgets for the info layout.
        """
        group_box = QGroupBox("Information")
        layout = QVBoxLayout()

        # Create a QLabel with explanatory text and hyperlink
        info_label = QLabel("Details on different hyperparameters can be found "
                            "<a href='https://docs.ultralytics.com/modes/train/#train-settings'>here</a>.")

        info_label.setOpenExternalLinks(True)
        info_label.setWordWrap(True)
        layout.addWidget(info_label)

        group_box.setLayout(layout)
        self.layout.addWidget(group_box)

    def setup_dataset_layout(self):
        raise NotImplementedError("Subclasses must implement this method.")

    def setup_model_layout(self):
        """
        Set up the layout and widgets for the model selection with a tabbed interface.
        """
        group_box = QGroupBox("Model Selection")
        layout = QVBoxLayout()

        # Create tabbed widget
        tab_widget = QTabWidget()

        # Tab 1: Select model from dropdown
        model_select_tab = QWidget()
        model_select_layout = QFormLayout(model_select_tab)

        # Model combo box
        self.model_combo = QComboBox()
        self.load_model_combobox()
        model_select_layout.addRow("Model:", self.model_combo)

        tab_widget.addTab(model_select_tab, "Select Model")

        # Tab 2: Use existing model
        model_existing_tab = QWidget()
        model_existing_layout = QFormLayout(model_existing_tab)

        # Existing Model
        self.model_edit = QLineEdit()
        self.model_button = QPushButton("Browse...")
        self.model_button.clicked.connect(self.browse_model_file)
        model_layout = QHBoxLayout()
        model_layout.addWidget(self.model_edit)
        model_layout.addWidget(self.model_button)
        model_existing_layout.addRow("Existing Model:", model_layout)

        tab_widget.addTab(model_existing_tab, "Use Existing Model")

        layout.addWidget(tab_widget)
        group_box.setLayout(layout)
        self.layout.addWidget(group_box)
        
    def setup_output_layout(self):
        """
        Set up the layout and widgets for the output directory.
        """
        group_box = QGroupBox("Output Parameters")
        form_layout = QFormLayout()

        # Project
        self.project_edit = QLineEdit()
        self.project_button = QPushButton("Browse...")
        self.project_button.clicked.connect(self.browse_project_dir)
        project_layout = QHBoxLayout()
        project_layout.addWidget(self.project_edit)
        project_layout.addWidget(self.project_button)
        form_layout.addRow("Project:", project_layout)

        # Name
        self.name_edit = QLineEdit()
        form_layout.addRow("Name:", self.name_edit)

        group_box.setLayout(form_layout)
        self.layout.addWidget(group_box)

    def setup_parameters_layout(self):
        """
        Set up the layout and widgets for the generic layout.
        """
        # Create helper function for boolean dropdowns
        def create_bool_combo():
            combo = QComboBox()
            combo.addItems(["True", "False"])
            return combo

        # Create parameters group box
        group_box = QGroupBox("Training Parameters")
        group_layout = QVBoxLayout(group_box)

        # Add import/export buttons at the top
        import_export_layout = QHBoxLayout()
        
        self.import_button = QPushButton("Import YAML")
        self.import_button.clicked.connect(self.import_parameters)
        import_export_layout.addWidget(self.import_button)

        self.export_button = QPushButton("Export YAML")
        self.export_button.clicked.connect(self.export_parameters)
        import_export_layout.addWidget(self.export_button)
        
        # Add stretch to push buttons to the left
        import_export_layout.addStretch()
        
        group_layout.addLayout(import_export_layout)

        # Create a widget to hold the form layout
        form_widget = QWidget()
        form_layout = QFormLayout(form_widget)

        # Create the scroll area
        scroll_area = QScrollArea()
        scroll_area.setWidgetResizable(True)
        scroll_area.setWidget(form_widget)
        
        group_layout.addWidget(scroll_area)

        # Create parameters
        # Epochs
        self.epochs_spinbox = QSpinBox()
        self.epochs_spinbox.setMinimum(1)
        self.epochs_spinbox.setMaximum(1000)
        self.epochs_spinbox.setValue(100)
        form_layout.addRow("Epochs:", self.epochs_spinbox)

        # Patience
        self.patience_spinbox = QSpinBox()
        self.patience_spinbox.setMinimum(1)
        self.patience_spinbox.setMaximum(1000)
        self.patience_spinbox.setValue(30)
        form_layout.addRow("Patience:", self.patience_spinbox)

        # Imgsz
        self.imgsz_spinbox = QSpinBox()
        self.imgsz_spinbox.setMinimum(16)
        self.imgsz_spinbox.setMaximum(4096)
        self.imgsz_spinbox.setValue(self.imgsz)
        form_layout.addRow("Image Size:", self.imgsz_spinbox)

        # Multi Scale
        self.multi_scale_combo = create_bool_combo()
        self.multi_scale_combo.setCurrentText("False")  # Default to False
        form_layout.addRow("Multi Scale:", self.multi_scale_combo)

        # Batch
        self.batch_spinbox = QSpinBox()
        self.batch_spinbox.setMinimum(1)
        self.batch_spinbox.setMaximum(1024)
        self.batch_spinbox.setValue(self.batch)
        form_layout.addRow("Batch Size:", self.batch_spinbox)

        # Workers
        self.workers_spinbox = QSpinBox()
        self.workers_spinbox.setMinimum(1)
        self.workers_spinbox.setMaximum(64)
        self.workers_spinbox.setValue(8)
        form_layout.addRow("Workers:", self.workers_spinbox)

        # Save
        self.save_combo = create_bool_combo()
        form_layout.addRow("Save:", self.save_combo)

        # Save Period
        self.save_period_spinbox = QSpinBox()
        self.save_period_spinbox.setMinimum(-1)
        self.save_period_spinbox.setMaximum(1000)
        self.save_period_spinbox.setValue(-1)
        form_layout.addRow("Save Period:", self.save_period_spinbox)

        # Freeze Layers
        self.freeze_layers_spinbox = QDoubleSpinBox()
        self.freeze_layers_spinbox.setMinimum(0.0)
        self.freeze_layers_spinbox.setMaximum(1.0)
        self.freeze_layers_spinbox.setSingleStep(0.01)
        self.freeze_layers_spinbox.setValue(0.0)
        form_layout.addRow("Freeze Layers:", self.freeze_layers_spinbox)

        # Weighted Dataset
        self.weighted_combo = create_bool_combo()
        form_layout.addRow("Weighted Sampling:", self.weighted_combo)

        # Dropout
        self.dropout_spinbox = QDoubleSpinBox()
        self.dropout_spinbox.setMinimum(0.0)
        self.dropout_spinbox.setMaximum(1.0)
        self.dropout_spinbox.setValue(0.0)
        form_layout.addRow("Dropout:", self.dropout_spinbox)

        # Optimizer
        self.optimizer_combo = QComboBox()
        self.optimizer_combo.addItems(["auto", "SGD", "Adam", "AdamW", "NAdam", "RAdam", "RMSProp"])
        self.optimizer_combo.setCurrentText("auto")
        form_layout.addRow("Optimizer:", self.optimizer_combo)

        # Val
        self.val_combo = create_bool_combo()
        form_layout.addRow("Validation:", self.val_combo)

        # Verbose
        self.verbose_combo = create_bool_combo()
        form_layout.addRow("Verbose:", self.verbose_combo)

        # Add horizontal separator line
        separator = QFrame()
        separator.setFrameShape(QFrame.HLine)
        separator.setFrameShadow(QFrame.Sunken)
        form_layout.addRow("", separator)
        
        # Add parameter button at the top of custom parameters section
        self.add_param_button = QPushButton("Add Parameter")
        self.add_param_button.clicked.connect(self.add_parameter_pair)
        form_layout.addRow("", self.add_param_button)

        # Add custom parameters section
        self.custom_params_layout = QVBoxLayout()
        form_layout.addRow("", self.custom_params_layout)

        # Remove parameter button at the bottom
        self.remove_param_button = QPushButton("Remove Parameter")
        self.remove_param_button.clicked.connect(self.remove_parameter_pair)
        self.remove_param_button.setEnabled(False)  # Disabled until at least one parameter is added
        form_layout.addRow("", self.remove_param_button)

        self.layout.addWidget(group_box)

    def add_parameter_pair(self):
        """
        Add a new parameter input group with name, value, and type selector.
        """
        param_layout = QHBoxLayout()

        # Parameter name field
        param_name = QLineEdit()
        param_name.setPlaceholderText("Parameter name")

        # Parameter value field
        param_value = QLineEdit()
        param_value.setPlaceholderText("Value")

        # Parameter type selector
        param_type = QComboBox()
        param_type.addItems(["string", "int", "float", "bool"])

        # Add widgets to layout
        param_layout.addWidget(param_name)
        param_layout.addWidget(param_value)
        param_layout.addWidget(param_type)

        # Store the widgets for later retrieval
        self.custom_params.append((param_name, param_value, param_type))
        self.custom_params_layout.addLayout(param_layout)

        # Enable the remove button since we now have at least one parameter
        self.remove_param_button.setEnabled(True)

    def remove_parameter_pair(self):
        """
        Remove the most recently added parameter pair.
        """
        if not self.custom_params:
            return

        # Get the last parameter group
        param_name, param_value, param_type = self.custom_params.pop()

        # Remove the layout containing these widgets
        layout_to_remove = self.custom_params_layout.itemAt(self.custom_params_layout.count() - 1)

        if layout_to_remove:
            # Remove and delete widgets from the layout
            while layout_to_remove.count():
                widget = layout_to_remove.takeAt(0).widget()
                if widget:
                    widget.deleteLater()

            # Remove the layout itself
            self.custom_params_layout.removeItem(layout_to_remove)

        # Disable the remove button if no more parameters
        if not self.custom_params:
            self.remove_param_button.setEnabled(False)

    def setup_buttons_layout(self):
        """

        """
        # Add OK and Cancel buttons
        self.buttons = QPushButton("OK")
        self.buttons.clicked.connect(self.accept)
        self.layout.addWidget(self.buttons)

        self.cancel_button = QPushButton("Cancel")
        self.cancel_button.clicked.connect(self.reject)
        self.layout.addWidget(self.cancel_button)

    def load_model_combobox(self):
        raise NotImplementedError("Subclasses must implement this method.")

    def browse_dataset_dir(self):
        """
        Browse and select a dataset directory.
        """
        dir_path = QFileDialog.getExistingDirectory(self, "Select Dataset Directory")
        if dir_path:
            # Load the class mapping if it exists
            class_mapping_path = f"{dir_path}/class_mapping.json"
            if os.path.exists(class_mapping_path):
                self.class_mapping = json.load(open(class_mapping_path, 'r'))
                self.mapping_edit.setText(class_mapping_path)

            # Set the dataset path
            self.dataset_edit.setText(dir_path)

    def browse_dataset_yaml(self):
        """
        Browse and select a dataset YAML file.
        """
        file_path, _ = QFileDialog.getOpenFileName(self,
                                                   "Select Dataset YAML File",
                                                   "",
                                                   "YAML Files (*.yaml *.yml)")
        if file_path:
            # Load the class mapping if it exists
            dir_path = os.path.dirname(file_path)
            class_mapping_path = f"{dir_path}/class_mapping.json"
            if os.path.exists(class_mapping_path):
                self.class_mapping = json.load(open(class_mapping_path, 'r'))
                self.mapping_edit.setText(class_mapping_path)

            # Set the dataset and class mapping paths
            self.dataset_edit.setText(file_path)

    def browse_class_mapping_file(self):
        """
        Browse and select a class mapping file.
        """
        file_path, _ = QFileDialog.getOpenFileName(self,
                                                   "Select Class Mapping File",
                                                   "",
                                                   "JSON Files (*.json)")
        if file_path:
            # Load the class mapping
            self.class_mapping = json.load(open(file_path, 'r'))

            # Set the class mapping path
            self.mapping_edit.setText(file_path)

    def browse_project_dir(self):
        """
        Browse and select a project directory.
        """
        dir_path = QFileDialog.getExistingDirectory(self, "Select Project Directory")
        if dir_path:
            self.project_edit.setText(dir_path)

    def browse_model_file(self):
        """
        Browse and select a model file.
        """
        file_path, _ = QFileDialog.getOpenFileName(self, "Select Model File")
        if file_path:
            self.model_edit.setText(file_path)

    def import_parameters(self):
        """
        Import parameters from a YAML file with automatic type inference.
        """
        file_path, _ = QFileDialog.getOpenFileName(self,
                                                   "Import Parameters from YAML",
                                                   "",
                                                   "YAML Files (*.yaml *.yml)")
        if not file_path:
            return

        try:
            with open(file_path, 'r') as f:
                data = yaml.safe_load(f)

            if not data:
                QMessageBox.warning(self, "Import Warning", "The YAML file appears to be empty or invalid.")
                return

            # For backward compatibility, check if the old nested 'parameters' key exists.
            # If not, use the whole data dictionary.
            params_to_load = data.get('parameters', data)

            # Helper function to infer type from value
            def infer_type_and_value(value):
                """
                Infer the type and convert the value based on its content.
                Returns (type_string, converted_value)
                """
                if isinstance(value, bool):
                    return "bool", value
                elif isinstance(value, int):
                    return "int", value
                elif isinstance(value, float):
                    return "float", value
                elif isinstance(value, str):
                    if value.lower() in ['true', 'false']:
                        return "bool", value.lower() == 'true'
                    try:
                        if '.' not in value:
                            return "int", int(value)
                        else:
                            return "float", float(value)
                    except ValueError:
                        return "string", value
                else:
                    return "string", str(value)

            # Clear existing custom parameters before importing
            while self.custom_params:
                self.remove_parameter_pair()

            # Map standard parameters to their UI controls
            param_mapping = {
                'epochs': self.epochs_spinbox,
                'patience': self.patience_spinbox,
                'imgsz': self.imgsz_spinbox,
                'multi_scale': self.multi_scale_combo,
                'batch': self.batch_spinbox,
                'workers': self.workers_spinbox,
                'save_period': self.save_period_spinbox,
                'freeze_layers': self.freeze_layers_spinbox,
                'dropout': self.dropout_spinbox,
                'save': self.save_combo,
                'weighted': self.weighted_combo,
                'val': self.val_combo,
                'verbose': self.verbose_combo,
                'optimizer': self.optimizer_combo
            }

            # Update UI controls with imported values
            for param_name, value in params_to_load.items():
                param_type, converted_value = infer_type_and_value(value)
                
                if param_name in param_mapping:
                    widget = param_mapping[param_name]
                    
                    if isinstance(widget, QSpinBox):
                        if isinstance(converted_value, (int, float)):
                            widget.setValue(int(converted_value))
                    elif isinstance(widget, QDoubleSpinBox):
                        if isinstance(converted_value, (int, float)):
                            widget.setValue(float(converted_value))
                    elif isinstance(widget, QComboBox):
                        if param_name in ['multi_scale', 'save', 'weighted', 'val', 'verbose']:
                            widget.setCurrentText("True" if converted_value else "False")
                        elif str(converted_value) in [widget.itemText(i) for i in range(widget.count())]:
                            widget.setCurrentText(str(converted_value))
                else:
                    # Add as a custom parameter
                    self.add_parameter_pair()
                    param_name_widget, param_value_widget, param_type_widget = self.custom_params[-1]
                    
                    param_name_widget.setText(param_name)
                    param_type_widget.setCurrentText(param_type)
                    
                    if param_type == "bool":
                        param_value_widget.setText("True" if converted_value else "False")
                    else:
                        param_value_widget.setText(str(converted_value))

            QMessageBox.information(self, 
                                    "Import Success", 
                                    "Parameters successfully imported with automatic type inference.")

        except Exception as e:
            QMessageBox.critical(self, "Import Error", f"Failed to import parameters: {str(e)}")

    def export_parameters(self):
        """
        Export current parameters to a flat YAML file.
        """
        file_path, _ = QFileDialog.getSaveFileName(self,
                                                   "Export Parameters to YAML",
                                                   "training_parameters.yaml",
                                                   "YAML Files (*.yaml *.yml)")
        if not file_path:
            return

        try:
            # Use a single flat dictionary for export
            export_data = {}

            # Standard parameters
            export_data['epochs'] = self.epochs_spinbox.value()
            export_data['patience'] = self.patience_spinbox.value()
            export_data['imgsz'] = self.imgsz_spinbox.value()
            export_data['batch'] = self.batch_spinbox.value()
            export_data['workers'] = self.workers_spinbox.value()
            export_data['save_period'] = self.save_period_spinbox.value()
            export_data['freeze_layers'] = self.freeze_layers_spinbox.value()
            export_data['dropout'] = self.dropout_spinbox.value()
            export_data['multi_scale'] = self.multi_scale_combo.currentText() == "True"
            export_data['save'] = self.save_combo.currentText() == "True"
            export_data['weighted'] = self.weighted_combo.currentText() == "True"
            export_data['val'] = self.val_combo.currentText() == "True"
            export_data['verbose'] = self.verbose_combo.currentText() == "True"
            export_data['optimizer'] = self.optimizer_combo.currentText()

            # Custom parameters
            for param_info in self.custom_params:
                param_name_widget, param_value_widget, param_type_widget = param_info
                name = param_name_widget.text().strip()
                value_str = param_value_widget.text().strip()
                type_name = param_type_widget.currentText()
                
                if name and value_str:
                    # Convert value to the correct type before exporting
                    try:
                        if type_name == "bool":
                            value = value_str.lower() == "true"
                        elif type_name == "int":
                            value = int(value_str)
                        elif type_name == "float":
                            value = float(value_str)
                        else:  # string type
                            value = value_str
                        export_data[name] = value
                    except ValueError:
                        # If conversion fails, save it as a string
                        print(f"Warning: Could not convert '{value_str}' to {type_name} for parameter '{name}'. "
                              "Saving as string.")
                        export_data[name] = value_str

            # Write the flat dictionary to the YAML file
            with open(file_path, 'w') as f:
                yaml.dump(export_data, f, default_flow_style=False, sort_keys=False, indent=2)

            QMessageBox.information(self, 
                                    "Export Success", 
                                    "Parameters successfully exported.")

        except Exception as e:
            QMessageBox.critical(self, 
                                 "Export Error", 
                                 f"Failed to export parameters: {str(e)}")

    def accept(self):
        """
        Handle the OK button click event.
        """
        self.train_model()
        super().accept()

    def get_parameters(self):
        """
        Get the training parameters from the dialog widgets.
    
        Returns:
            dict: A dictionary of training parameters.
        """
        # Extract values from dialog widgets
        params = {
            'task': self.task,
            'project': self.project_edit.text(),
            'name': self.name_edit.text(),
            'data': self.dataset_edit.text(),
            'epochs': self.epochs_spinbox.value(),
            'patience': self.patience_spinbox.value(),
            'batch': self.batch_spinbox.value(),
            'imgsz': self.imgsz_spinbox.value(),
            'multi_scale': self.multi_scale_combo.currentText() == "True",
            'save': self.save_combo.currentText() == "True",
            'save_period': self.save_period_spinbox.value(),
            'workers': self.workers_spinbox.value(),
            'optimizer': self.optimizer_combo.currentText(),
            'verbose': self.verbose_combo.currentText() == "True",
            'freeze_layers': self.freeze_layers_spinbox.value(),
            'weighted': self.weighted_combo.currentText() == "True",
            'dropout': self.dropout_spinbox.value(),
            'val': self.val_combo.currentText() == "True",
            'exist_ok': True,
            'plots': True,
        }
        # Default project folder
        project = 'Data/Training'
        params['project'] = params['project'] if params['project'] else project
        # Default project name
        now = datetime.datetime.now()
        now = now.strftime("%Y-%m-%d_%H-%M-%S")
        params['name'] = params['name'] if params['name'] else now
        # Either the model path, or the model name provided from combo box
        params['model'] = self.model_edit.text() if self.model_edit.text() else self.model_combo.currentText()
    
        # Add custom parameters (allows overriding the above parameters)
        for param_info in self.custom_params:
            param_name, param_value, param_type = param_info
            name = param_name.text().strip()
            value = param_value.text().strip()
            type_name = param_type.currentText()
            
            if name:
                if type_name == "bool":
                    params[name] = value.lower() == "true"
                elif type_name == "int":
                    try:
                        params[name] = int(value)
                    except ValueError:
                        print(f"Warning: Could not convert '{value}' to int for parameter '{name}', using as string")
                        params[name] = value
                elif type_name == "float":
                    try:
                        params[name] = float(value)
                    except ValueError:
                        print(f"Warning: Could not convert '{value}' to float for parameter '{name}', using as string")
                        params[name] = value
                else:  # string type
                    params[name] = value
    
        # Return the dictionary of parameters
        return params

    def train_model(self):
        """
        Train the model based on the provided parameters.
        """
        # Get training parameters
        self.params = self.get_parameters()

        # Create and start the worker thread
        self.worker = TrainModelWorker(self.params, self.main_window.device)
        self.worker.training_started.connect(self.on_training_started)
        self.worker.training_completed.connect(self.on_training_completed)
        self.worker.training_error.connect(self.on_training_error)
        self.worker.start()

    def on_training_started(self):
        """
        Handle the event when the training starts.
        """
        # Save the class mapping JSON file
        output_dir_path = os.path.join(self.params['project'], self.params['name'])
        os.makedirs(output_dir_path, exist_ok=True)
        if self.class_mapping:
            # Write the json file to the output directory
            with open(f"{output_dir_path}/class_mapping.json", 'w') as json_file:
                json.dump(self.class_mapping, json_file, indent=4)

        message = "Model training has commenced.\nMonitor the console for real-time progress."
        QMessageBox.information(self, "Model Training Status", message)
        
    def on_training_error(self, error_message):
        """
        Handle the event when an error occurs during training.

        Args:
            error_message (str): The error message.
        """
        QMessageBox.critical(self, "Error", error_message)
        print(error_message)

    def on_training_completed(self):
        """
        Handle the event when the training completes.
        """
        msg_box = QMessageBox(self)
        msg_box.setWindowTitle("Model Training Status")
        msg_box.setText("Model training has successfully been completed.")
        ok_button = msg_box.addButton(QMessageBox.Ok)
        
        deploy_button = msg_box.addButton("Deploy Model", QMessageBox.AcceptRole)
        msg_box.exec_()

        if msg_box.clickedButton() == deploy_button:
            self.deploy_trained_model()

    def deploy_trained_model(self):
        """
        Load the trained model and class mapping into the existing deployment dialog,
        update the status and labels, but do not show the dialog.
        """
        output_folder = f"{self.params['project']}/{self.params['name']}"
        weights_folder = f"{output_folder}/weights"
        
        # Find the best weights file (usually 'best.pt' or similar)
        best_weights = None
        for fname in os.listdir(weights_folder):
            if fname.startswith("best") and fname.endswith(".pt"):
                best_weights = f"{weights_folder}/{fname}"
                break

        if not best_weights:
            QMessageBox.warning(self, "Deploy Model", "Could not find trained model weights.")
            return

        # Load class mapping if available
        class_mapping_path = f"{output_folder}/class_mapping.json"
        class_mapping = {}
        if os.path.exists(class_mapping_path):
            try:
                with open(class_mapping_path, "r") as f:
                    class_mapping = json.load(f)
            except Exception as e:
                QMessageBox.warning(self, "Deploy Model", f"Failed to load class mapping: {str(e)}")

        # Use the existing deployment dialog instance from main_window
        if self.task == "classify":
            deploy_dialog = self.main_window.classify_deploy_model_dialog
        elif self.task == "detect":
            deploy_dialog = self.main_window.detect_deploy_model_dialog
        elif self.task == "segment":
            deploy_dialog = self.main_window.segment_deploy_model_dialog
        else:
            QMessageBox.warning(self, "Deploy Model", "Unknown task type for deployment.")
            return

        # Set the model path and class mapping, then load the model
        deploy_dialog.model_path = best_weights
        deploy_dialog.class_mapping = class_mapping
        deploy_dialog.load_model()

        # Update label window and status bar
        if hasattr(deploy_dialog, "add_labels_to_label_window"):
            deploy_dialog.add_labels_to_label_window()
        if hasattr(deploy_dialog, "check_and_display_class_names"):
            deploy_dialog.check_and_display_class_names()
        if hasattr(deploy_dialog, "status_bar"):
            deploy_dialog.status_bar.setText(f"Model loaded: {os.path.basename(best_weights)}")

    # Do NOT show the dialog (no exec_())





from PyQt5.QtCore import Qt, QPointF
from PyQt5.QtWidgets import QMessageBox

from coralnet_toolbox.Tools.QtTool import Tool


# ----------------------------------------------------------------------------------------------------------------------
# Classes
# ----------------------------------------------------------------------------------------------------------------------


class FillTool(Tool):
    """A tool for filling contiguous regions in a MaskAnnotation layer with the selected label."""
    def __init__(self, annotation_window):
        # Call the parent constructor to set up annotation_window, etc.
        super().__init__(annotation_window)
        
        # Disable crosshair for this tool
        self.show_crosshair = True
        
        # You can set a specific cursor for this tool
        self.cursor = Qt.CrossCursor

    def mousePressEvent(self, event):
        """Handles left-click to fill the region under the cursor."""
        if event.button() != Qt.LeftButton:
            return
            
        if not self.annotation_window.selected_label:
            QMessageBox.warning(self.annotation_window,
                                "No Label Selected",
                                "A label must be selected before using the fill tool.")
            return
        
        if not self.annotation_window.cursorInWindow(event.pos()):
            return
        
        self._apply_fill(event)

    def mouseMoveEvent(self, event):
        """Handles mouse movement, shows crosshair if enabled."""
        # Call the parent method to handle crosshair
        super().mouseMoveEvent(event)
    
    def mouseReleaseEvent(self, event):
        """Called when the mouse is released."""
        pass  # No action needed on release

    def wheelEvent(self, event):
        """No wheel event handling for fill tool."""
        pass

    def create_cursor_annotation(self, scene_pos: QPointF = None):
        """No special cursor annotation for fill tool."""
        pass

    def update_cursor_annotation(self, scene_pos: QPointF = None):
        """No update needed."""
        pass

    def clear_cursor_annotation(self):
        """No cursor annotation to clear."""
        pass

    def stop_current_drawing(self):
        """No drawing to stop."""
        pass

    def _apply_fill(self, event):
        """Fills the contiguous region under the cursor with the selected label."""
        # Get the current mask annotation from the annotation window
        mask_annotation = self.annotation_window.current_mask_annotation
        
        # Ensure that a mask and a selected label are active
        if not mask_annotation or not self.annotation_window.selected_label:
            return
            
        # Get the mouse position in the scene's coordinate system
        scene_pos = self.annotation_window.mapToScene(event.pos())
        
        # Get the selected label's class ID
        selected_label_id = self.annotation_window.selected_label.id
        new_class_id = mask_annotation.label_id_to_class_id_map.get(selected_label_id)
        
        if new_class_id is None:
            return  # Label not found in map
        
        # Call the fill_region method on the MaskAnnotation object
        mask_annotation.fill_region(scene_pos, new_class_id)
        
        # Update the display to reflect changes
        self.annotation_window.update_scene()



from .QtPanTool import PanTool
from .QtPatchTool import PatchTool
from .QtRectangleTool import RectangleTool
from .QtPolygonTool import PolygonTool

from .QtBrushTool import BrushTool
from .QtEraseTool import EraseTool
from .QtFillTool import FillTool

from .QtSAMTool import SAMTool
from .QtSeeAnythingTool import SeeAnythingTool
from .QtSelectTool import SelectTool 
from .QtZoomTool import ZoomTool
from .QtWorkAreaTool import WorkAreaTool

from .QtCutSubTool import CutSubTool
from .QtMoveSubTool import MoveSubTool
from .QtResizeSubTool import ResizeSubTool
from .QtSelectSubTool import SelectSubTool
from .QtSubtractSubTool import SubtractSubTool

__all__ = [
    'PanTool',
    'PatchTool',
    'RectangleTool',
    'PolygonTool', 
    'BrushTool',
    'EraseTool',
    'FillTool',
    'SAMTool',
    'SeeAnythingTool',
    'SelectTool',
    'ZoomTool',
    'WorkAreaTool',
    'CutSubTool',
    'MoveSubTool',
    'ResizeSubTool',
    'SelectSubTool',
    'SubtractSubTool',
]


import warnings

import base64
import rasterio

import numpy as np

from scipy.ndimage import label as ndimage_label
from skimage.measure import find_contours

from shapely.geometry import Polygon
from rasterio.features import rasterize
from rasterio.transform import from_origin
from pycocotools import mask

from PyQt5.QtCore import Qt, QPointF, QRectF
from PyQt5.QtWidgets import QGraphicsScene, QGraphicsPixmapItem, QGraphicsItem, QApplication
from PyQt5.QtGui import QPixmap, QColor, QImage, QPainter, QBrush, QPolygonF

from coralnet_toolbox.Annotations.QtAnnotation import Annotation
from coralnet_toolbox.Annotations.QtPolygonAnnotation import PolygonAnnotation

warnings.filterwarnings("ignore", category=DeprecationWarning)


# ----------------------------------------------------------------------------------------------------------------------
# Classes
# ----------------------------------------------------------------------------------------------------------------------


class MaskGraphicsItem(QGraphicsItem):
    def __init__(self, mask_annotation):
        super().__init__()
        self.mask_annotation = mask_annotation
        self.setFlag(QGraphicsItem.ItemUsesExtendedStyleOption, True)

    def boundingRect(self):
        height, width = self.mask_annotation.mask_data.shape
        return QRectF(0, 0, width, height)

    def paint(self, painter, option, widget):
        if self.mask_annotation.qimage:
            # Apply transparency at render time for instant updates
            transparency = self.mask_annotation.get_current_transparency()
            if transparency < 255:
                painter.setOpacity(transparency / 255.0)
            painter.drawImage(0, 0, self.mask_annotation.qimage)
            if transparency < 255:
                painter.setOpacity(1.0)  # Reset opacity for other drawing operations


class MaskAnnotation(Annotation):
    LOCK_BIT = 2**7  # For uint8, this is 128

    def __init__(self,
                 image_path: str,
                 mask_data: np.ndarray,
                 initial_labels: list,
                 transparency: int = 128,
                 show_msg: bool = False,
                 rasterio_src=None):
        """
        Initialize a full-image semantic segmentation annotation.
        There should only be one MaskAnnotation per image.
        """
        # --- This block is for context ---
        if not initial_labels:
            raise ValueError("initial_labels cannot be empty.")
        placeholder_label = initial_labels[0]

        super().__init__(
            short_label_code=placeholder_label.short_label_code,
            long_label_code=placeholder_label.long_label_code,
            color=placeholder_label.color,
            image_path=image_path,
            label_id=placeholder_label.id,
            transparency=transparency,
            show_msg=show_msg
        )
        
        self.mask_data = mask_data.astype(np.uint8)
        
        self.class_id_to_label_map = {}
        self.label_id_to_class_id_map = {}
        self.visible_label_ids = set()
        
        self.next_class_id = 1
        self.sync_label_map(initial_labels)
        # Update visible labels to include all labels currently in the map
        self.visible_label_ids = set(self.label_id_to_class_id_map.keys())

        self.offset = QPointF(0, 0)
        self.rasterio_src = rasterio_src
        
        # Initialize the colored canvas and QImage once at creation time.
        # This moves the expensive, one-time image generation from the first
        # brush stroke to the object's creation.
        self.colored_mask = None
        self.qimage = None
        self._initialize_canvas()
        
        self.set_centroid()
        self.set_cropped_bbox()

    def sync_label_map(self, current_labels_in_project: list):
        """Ensures the internal maps are synced with the project's labels,
           assigning new, stable IDs to any new labels without changing existing ones."""
        
        # Add any new labels from the project list to our maps
        for label in current_labels_in_project:
            if label.id not in self.label_id_to_class_id_map:
                new_id = self.next_class_id
                self.class_id_to_label_map[new_id] = label
                self.label_id_to_class_id_map[label.id] = new_id
                self.next_class_id += 1
                
                # Also add the new label's ID to the set of visible labels.
                self.visible_label_ids.add(label.id)
                
    def _build_color_map(self):
        """Builds a numpy array mapping class IDs to RGBA colors."""
        # Ensure the map is large enough to handle locked class IDs
        max_id = max(self.class_id_to_label_map.keys()) if self.class_id_to_label_map else 0
        map_size = max(256, max_id + self.LOCK_BIT + 1)
        color_map = np.zeros((map_size, 4), dtype=np.uint8)

        for class_id, label in self.class_id_to_label_map.items():
            color = label.color
            # Always use full alpha in the array - transparency applied at render time
            alpha = 255 if label.id in self.visible_label_ids else 0
            
            # Set the color for both the normal and the locked version of the class ID
            rgba = [color.red(), color.green(), color.blue(), alpha]
            color_map[class_id] = rgba
            if class_id + self.LOCK_BIT < map_size:
                color_map[class_id + self.LOCK_BIT] = rgba

        return color_map
    
    def _initialize_canvas(self):
        """
        Creates the initial color canvas and QImage. This is the one-time expensive
        operation that happens when the mask is first loaded.
        """
        height, width = self.mask_data.shape
        color_map = self._build_color_map()
        
        # Create the full-size 4-channel RGBA numpy array
        self.colored_mask = color_map[self.mask_data]
        
        # Create a QImage that is a VIEW on the numpy array's data buffer.
        # Modifying the numpy array will now automatically update the QImage.
        self.qimage = QImage(self.colored_mask.data, width, height, QImage.Format_RGBA8888)

    def _update_full_canvas(self):
        """Regenerates the entire color canvas. Used for global changes like label color edits."""
        color_map = self._build_color_map()
        # Update the existing colored_mask array in-place
        np.copyto(self.colored_mask, color_map[self.mask_data])

    def _update_canvas_slice(self, update_rect):
        """
        Efficiently updates only a small rectangular slice of the color canvas.
        Used for tools like the brush.
        """
        x1, y1, x2, y2 = update_rect
        
        # Get the slice of the data that has changed
        data_slice = self.mask_data[y1:y2, x1:x2]
        
        # Rebuild the color map in case label properties have changed
        color_map = self._build_color_map()
        
        # Perform the color lookup ONLY for the small data slice
        color_slice = color_map[data_slice]
        
        # Update the corresponding slice in our persistent color canvas
        self.colored_mask[y1:y2, x1:x2] = color_slice

    def set_centroid(self):
        """Set the centroid to the center of the image."""
        height, width = self.mask_data.shape
        self.center_xy = QPointF(width / 2.0, height / 2.0)

    def set_cropped_bbox(self):
        """Set the bounding box to the full dimensions of the image."""
        height, width = self.mask_data.shape
        self.cropped_bbox = (0, 0, width, height)
        self.annotation_size = int(max(width, height))
                
    def contains_point(self, point: QPointF) -> bool:
        """Check if a point is within the mask's classified area."""
        x, y = int(point.x()), int(point.y())
        height, width = self.mask_data.shape
        if 0 <= y < height and 0 <= x < width:
            return self.mask_data[y, x] > 0
        return False

    def get_area(self):
        """Return the total number of non-background pixels."""
        return np.count_nonzero(self.mask_data)

    def get_bounding_box_top_left(self):
        """Get the top-left corner of the annotation's bounding box (always 0,0)."""
        return QPointF(0, 0)

    def get_bounding_box_bottom_right(self):
        """Get the bottom-right corner of the annotation's bounding box."""
        height, width = self.mask_data.shape
        return QPointF(width, height)

    def create_graphics_item(self, scene: QGraphicsScene):
        """Create a QGraphicsPixmapItem to display the mask."""
        self.graphics_item = MaskGraphicsItem(self)
        scene.addItem(self.graphics_item)

    def update_graphics_item(self, update_rect=None):
        """Update the graphics item when mask data has changed."""
        if self.graphics_item is None:
            return
        
        if update_rect:
            # Localized update for brush strokes - update only the changed area
            self._update_canvas_slice(update_rect)
            qt_rect = QRectF(update_rect[0], 
                             update_rect[1], 
                             update_rect[2] - update_rect[0], 
                             update_rect[3] - update_rect[1])
            self.graphics_item.update(qt_rect)
        else:
            # Full update for global changes (e.g., label color changes)
            self._update_full_canvas()
            self.graphics_item.update()

    def update_mask(self, brush_location: QPointF, brush_mask: np.ndarray, new_class_id: int):
        """
        Modify the mask data based on a brush stroke, respecting pre-locked pixels.
        """
        x_start, y_start = int(brush_location.x()), int(brush_location.y())
        brush_h, brush_w = brush_mask.shape
        mask_h, mask_w = self.mask_data.shape

        # Define the update area and clip it to the mask's bounds
        x_end = min(x_start + brush_w, mask_w)
        y_end = min(y_start + brush_h, mask_h)
        clipped_x_start = max(x_start, 0)
        clipped_y_start = max(y_start, 0)

        if clipped_x_start >= x_end or clipped_y_start >= y_end:
            return
            
        # Get the slice of the main mask data we will be updating
        target_slice = self.mask_data[clipped_y_start:y_end, clipped_x_start:x_end]
        
        # Clip the user's brush mask to match the on-screen portion
        brush_x_offset = clipped_x_start - x_start
        brush_y_offset = clipped_y_start - y_start
        clipped_brush_mask = brush_mask[brush_y_offset:brush_y_offset + target_slice.shape[0],
                                        brush_x_offset:brush_x_offset + target_slice.shape[1]]
        
        # Identify pixels within the target slice that are NOT locked
        unlocked_pixels_mask = target_slice < self.LOCK_BIT
        
        # The final brush can only be applied where the brush shape is true AND the target pixel is not locked
        final_brush_mask = clipped_brush_mask & unlocked_pixels_mask

        # Apply the final, protected brush mask to the data
        target_slice[final_brush_mask] = new_class_id
        
        # Trigger a visual update for the changed rectangle
        changed_rect_coords = (clipped_x_start, clipped_y_start, x_end, y_end)
        self.update_graphics_item(update_rect=changed_rect_coords)
        
        self.annotationUpdated.emit(self)
        
    def get_current_transparency(self):
        """Get the current transparency value for rendering."""
        # Try to get the active label's transparency from the application
        try:
            from PyQt5.QtWidgets import QApplication
            app = QApplication.instance()
            if app:
                main_window = app.activeWindow()
                if hasattr(main_window, 'label_window') and hasattr(main_window.label_window, 'active_label'):
                    active_label = main_window.label_window.active_label
                    if active_label:
                        return active_label.transparency
        except Exception:
            pass
        return self.transparency
    
    def update_transparency(self, transparency):
        """Update transparency instantly using render-time application."""
        transparency = max(0, min(255, transparency))  # Clamp to valid range
        if self.transparency != transparency:
            self.transparency = transparency
            # Update transparency on all visible labels
            for label_id in self.visible_label_ids:
                label = next((lbl for lbl in self.class_id_to_label_map.values() if lbl.id == label_id), None)
                if label:
                    label.transparency = transparency
            
            # Trigger repaint - transparency applied during rendering
            if self.graphics_item:
                self.graphics_item.update()
            
    def update_visible_labels(self, visible_ids: set):
        """
        Updates the set of visible label IDs and triggers a redraw of the mask.

        Args:
            visible_ids (set): A set containing the UUIDs of labels that should be visible.
        """
        self.visible_label_ids = visible_ids
        self.update_graphics_item()
            
    def remove_from_scene(self):
        """Removes the graphics item from its scene, if it exists."""
        if self.graphics_item and self.graphics_item.scene():
            self.graphics_item.scene().removeItem(self.graphics_item)
            
        # Remove the graphics item reference
        self.graphics_item = None

    # --- Data Manipulation & Editing Methods ---

    def fill_region(self, point: QPointF, new_class_id: int):
        """
        Fills a contiguous region with a new class ID, respecting pre-locked pixels.
        """
        x, y = int(point.x()), int(point.y())
        height, width = self.mask_data.shape
        if not (0 <= y < height and 0 <= x < width):
            return

        # Check if the starting pixel is locked. If so, we cannot fill from it.
        if self.mask_data[y, x] >= self.LOCK_BIT:
            return

        old_class_id = self.mask_data[y, x]
        if old_class_id == new_class_id:
            return
        
        QApplication.setOverrideCursor(Qt.WaitCursor)

        # Find all contiguous pixels with the same starting class ID. This automatically respects
        # locked areas because they will have a different ID (e.g., 3 vs 131) and act as a border.
        labeled_array, num_features = ndimage_label(self.mask_data == old_class_id)
        region_label = labeled_array[y, x]
        fill_mask = labeled_array == region_label
        
        # Apply the fill. No further checks are needed because locked pixels were excluded by the ndimage_label step.
        self.mask_data[fill_mask] = new_class_id
        
        # Use localized update for efficiency
        coords = np.where(fill_mask)
        y_min, y_max = coords[0].min(), coords[0].max()
        x_min, x_max = coords[1].min(), coords[1].max()
        update_rect = (x_min, y_min, x_max + 1, y_max + 1)
        self.update_graphics_item(update_rect=update_rect)

        QApplication.restoreOverrideCursor()
        self.annotationUpdated.emit(self)

    def replace_class(self, old_class_id: int, new_class_id: int):
        """Replaces all pixels of one class ID with another across the entire mask."""
        self.mask_data[self.mask_data == old_class_id] = new_class_id
        self.update_graphics_item()
        self.annotationUpdated.emit(self)

    def _rasterize_annotation_group(self, annotations: list, class_id: int):
        """Burns a list of vector annotations OF THE SAME CLASS into the mask."""
        height, width = self.mask_data.shape

        geometries = []
        for anno in annotations:
            if hasattr(anno, 'get_polygon'):
                qt_polygon = anno.get_polygon()
                points = [(p.x(), p.y()) for p in qt_polygon]
                if len(points) >= 3:
                    geometries.append(Polygon(points))

        if not geometries:
            return

        # Rasterize all geometries into a single boolean mask
        stencil_np = rasterize(
            geometries,
            out_shape=(height, width),
            transform=from_origin(0, 0, 1, 1),
            fill=0,          # Pixels outside shapes are 0 (not burned)
            default_value=1, # Pixels inside shapes are 1 (burned)
            dtype=np.uint8
        ).astype(bool)

        # Add the LOCK_BIT to the class_id to mark these pixels as protected.
        self.mask_data[stencil_np] = class_id + self.LOCK_BIT
        
    def _create_stencil_for_group(self, annotations: list) -> np.ndarray:
        """Creates a boolean numpy stencil for a list of annotations using rasterio."""
        height, width = self.mask_data.shape

        geometries = []
        for anno in annotations:
            if hasattr(anno, 'get_polygon'):
                qt_polygon = anno.get_polygon()
                points = [(p.x(), p.y()) for p in qt_polygon]
                if len(points) >= 3:
                    geometries.append(Polygon(points))

        if not geometries:
            return np.zeros((height, width), dtype=bool)

        # Rasterize all geometries into a single boolean mask
        stencil = rasterize(
            geometries,
            out_shape=(height, width),
            transform=from_origin(0, 0, 1, 1),
            fill=0,          # Pixels outside shapes are 0 (not in stencil)
            default_value=1, # Pixels inside shapes are 1 (in stencil)
            dtype=np.uint8
        ).astype(bool)
        
        return stencil
    
    def _create_local_lock_mask(self, rect: QRectF, annotations: list) -> np.ndarray:
        """
        Creates a boolean numpy stencil for a given rectangular area by rasterizing
        the provided annotations. True means the pixel is protected.
        """
        width, height = int(rect.width()), int(rect.height())
        if width <= 0 or height <= 0:
            return np.zeros((height, width), dtype=bool)

        geometries = []
        for anno in annotations:
            if hasattr(anno, 'get_polygon'):
                qt_polygon = anno.get_polygon()
                points = [(p.x(), p.y()) for p in qt_polygon]
                if len(points) >= 3:
                    geometries.append(Polygon(points))

        if not geometries:
            return np.zeros((height, width), dtype=bool)

        # Create a transform that maps the global coordinates of the geometries
        # to the local coordinate system of our output numpy array.
        transform = from_origin(rect.x(), rect.y(), 1, 1)

        # Rasterize the shapes into a numpy array.
        lock_mask = rasterize(
            geometries,
            out_shape=(height, width),
            transform=transform,
            fill=0,          # Pixels outside the shapes are 0 (not locked)
            default_value=1, # Pixels inside the shapes are 1 (locked)
            dtype=np.uint8
        ).astype(bool)
        
        return lock_mask
    
    def _create_lock_mask_with_rasterio(self, rect: QRectF, annotations: list) -> np.ndarray:
        """
        Creates a boolean lock mask for a given rectangular area using the
        highly optimized rasterio.features.rasterize function.
        """
        geometries = []
        for anno in annotations:
            if hasattr(anno, 'get_polygon'):
                qt_polygon = anno.get_polygon()
                points = [(p.x(), p.y()) for p in qt_polygon]
                if len(points) >= 3:
                    geometries.append(Polygon(points))
        
        if not geometries:
            return np.zeros((int(rect.height()), int(rect.width())), dtype=bool)
    
        # Create a transform that maps the global coordinates of the geometries
        # to the local coordinate system of our output numpy array.
        transform = from_origin(rect.x(), rect.y(), 1, 1)
    
        # Rasterize the shapes into a numpy array.
        lock_mask = rasterize(
            geometries,
            out_shape=(int(rect.height()), int(rect.width())),
            transform=transform,
            fill=0,          # Pixels outside the shapes are 0 (not locked)
            default_value=1, # Pixels inside the shapes are 1 (locked)
            dtype=np.uint8
        ).astype(bool)
        
        return lock_mask

    def rasterize_annotations(self, all_annotations: list):
        """
        Mark pixels covered by vector annotations as locked to prevent painting over them.
        Uses rasterio.features.rasterize for robust and efficient polygon rasterization.
        Vector annotations remain visible while their pixel areas become protected.
        
        Args:
            all_annotations: List of vector annotations to protect
        """
        if not all_annotations:
            return

        height, width = self.mask_data.shape

        geometries = []
        for annotation in all_annotations:
            if not hasattr(annotation, 'get_polygon'):
                continue
            try:
                polygon = annotation.get_polygon()
                if polygon is None or polygon.isEmpty():
                    continue
                points = [(polygon.at(i).x(), polygon.at(i).y()) for i in range(polygon.count())]
                if len(points) < 3:
                    continue
                geometries.append(Polygon(points))
            except Exception as e:
                print(f"Warning: Could not process annotation {annotation.id}: {e}")
                continue

        if not geometries:
            return

        # Rasterize all geometries into a single boolean mask
        lock_mask = rasterize(
            geometries,
            out_shape=(height, width),
            transform=from_origin(0, 0, 1, 1),
            fill=0,           # Pixels outside shapes are 0 (not locked)
            default_value=1,  # Pixels inside shapes are 1 (locked)
            dtype=np.uint8
        ).astype(bool)

        # Apply locking: Add LOCK_BIT to pixels that are not already locked
        to_lock = lock_mask & (self.mask_data < self.LOCK_BIT)
        self.mask_data[to_lock] += self.LOCK_BIT

    def unrasterize_annotations(self):
        """
        Remove lock protection from all pixels that were marked as locked.
        This allows mask editing over previously protected vector annotation areas.
        """
        # Find all pixels that have the lock bit set and remove it
        locked_pixels = self.mask_data >= self.LOCK_BIT
        
        # Remove the lock bit from these pixels, keeping their original class
        self.mask_data[locked_pixels] = self.mask_data[locked_pixels] - self.LOCK_BIT

    def clear_pixels_for_class(self, class_id: int):
        """Finds all pixels matching a class ID (both locked and unlocked) and resets them to 0."""
        if class_id == 0:  # Cannot clear background class
            return

        # Create a boolean mask of all pixels whose real class ID matches.
        pixels_to_clear = (self.mask_data % self.LOCK_BIT) == class_id
        
        # Set these pixels back to 0 (unclassified).
        self.mask_data[pixels_to_clear] = 0
        
    def clear_pixels_for_annotations(self, annotations_to_clear: list):
        """
        Rasterizes a list of vector annotations and sets the corresponding
        pixels in the mask_data to 0 (unclassified).
        """
        if not annotations_to_clear:
            return

        # 1. Convert all annotation polygons to Shapely Polygons.
        geometries = []
        for anno in annotations_to_clear:
            if hasattr(anno, 'get_polygon'):
                qt_polygon = anno.get_polygon()
                points = [(p.x(), p.y()) for p in qt_polygon]
                if len(points) >= 3:
                    geometries.append(Polygon(points))

        if not geometries:
            return

        # 2. Rasterize all shapes at once into a boolean mask.
        # This creates a numpy array where 'True' indicates a pixel is covered
        # by at least one of the vector annotations.
        height, width = self.mask_data.shape
        clear_mask = rasterize(
            geometries,
            out_shape=(height, width),
            fill=0,
            default_value=1,
            dtype=np.uint8
        ).astype(bool)

        # 3. Apply the mask to the data, setting pixels to 0.
        self.mask_data[clear_mask] = 0

        # 4. Trigger a full repaint of the mask to show the changes.
        self.update_graphics_item()

    # --- Analysis & Information Retrieval Methods ---

    def get_class_statistics(self) -> dict:
        """Returns a dictionary with pixel counts and percentages for each class."""
        stats = {}
        total_pixels = self.get_area()
        if total_pixels == 0:
            return stats

        class_ids, counts = np.unique(self.mask_data[self.mask_data > 0], return_counts=True)
        
        for cid, count in zip(class_ids, counts):
            label = self.class_id_to_label_map.get(int(cid))
            if label:
                stats[label.short_label_code] = {
                    "pixel_count": int(count),
                    "percentage": (count / total_pixels) * 100
                }
        return stats

    def get_class_at_point(self, point: QPointF) -> int:
        """Returns the class ID at a specific point."""
        x, y = int(point.x()), int(point.y())
        height, width = self.mask_data.shape
        if 0 <= y < height and 0 <= x < width:
            return self.mask_data[y, x]
        return 0  # Return background class if outside bounds

    # --- Conversion & Exporting Methods ---
    
    def get_binary_mask(self, class_id: int) -> np.ndarray:
        """Returns a boolean numpy array where True corresponds to the given class ID."""
        return self.mask_data == class_id

    def to_instance_polygons(self, class_id: int) -> list:
        """Converts all contiguous regions of a class ID into PolygonAnnotations."""
        binary_mask = self.get_binary_mask(class_id)
        # Add padding to handle contours touching the border
        padded_mask = np.pad(binary_mask, pad_width=1, mode='constant', constant_values=0)
        
        # Level is 0.5 to find contours between 0 and 1 values
        contours = find_contours(padded_mask, level=0.5)
        
        annotations = []
        for contour in contours:
            # Remove padding offset and swap (row, col) to (x, y)
            points = [QPointF(p[1] - 1, p[0] - 1) for p in contour]
            if len(points) > 2:  # Must have at least 3 points for a valid polygon
                # Use the label associated with the class_id for the new annotation
                label = self.class_id_to_label_map[class_id]
                anno = PolygonAnnotation(
                    points=points,
                    short_label_code=label.short_label_code,
                    long_label_code=label.long_label_code,
                    color=label.color,
                    image_path=self.image_path,
                    label_id=label.id
                )
                annotations.append(anno)
        return annotations

    def export_as_png(self, path: str, use_label_colors: bool = True):
        """Saves the mask to a PNG file."""
        if use_label_colors:
            # Use current colored image for export
            self.qimage.save(path)
        else:
            # Save the raw class IDs as a grayscale image
            height, width = self.mask_data.shape
            # Ensure data is in a format QImage can handle (e.g., 8-bit grayscale)
            if self.mask_data.max() < 256:
                img_data = self.mask_data.astype(np.uint8)
                q_image = QImage(img_data.data, width, height, QImage.Format_Grayscale8)
                q_image.save(path)
            else:
                warnings.warn("Mask contains class IDs > 255; cannot save as 8-bit grayscale PNG.")

    def export_as_raster(self, path: str):
        """Saves the mask data to a raster file (e.g., GeoTIFF) using rasterio."""
        profile = {
            'driver': 'GTiff',
            'height': self.mask_data.shape[0],
            'width': self.mask_data.shape[1],
            'count': 1,
            'dtype': self.mask_data.dtype
        }
        
        # If the original image was opened with rasterio, copy its spatial metadata
        if self.rasterio_src:
            profile['crs'] = self.rasterio_src.crs
            profile['transform'] = self.rasterio_src.transform

        with rasterio.open(path, 'w', **profile) as dst:
            dst.write(self.mask_data, 1)

    # --- Serialization & Deserialization ---

    def to_dict(self):
        """Serialize the annotation to a dictionary, with RLE for the mask."""
        base_dict = super().to_dict()
        
        # Encode each class's binary mask using pycocotools
        rle_list = []
        unique_classes = np.unique(self.mask_data)
        for class_id in unique_classes:
            if class_id == 0:
                continue
            binary_mask = (self.mask_data == class_id).astype(np.uint8)
            rle = mask.encode(np.asfortranarray(binary_mask))
            rle['counts'] = base64.b64encode(rle['counts']).decode('ascii')
            rle_list.append({'class_id': int(class_id), 'rle': rle})
        
        # Convert the label map to a serializable format
        serializable_label_map = {}
        for cid, label in self.class_id_to_label_map.items():
            serializable_label_map[cid] = label.short_label_code

        base_dict.update({
            'shape': self.mask_data.shape,
            'rle_masks': rle_list,
            'label_map': serializable_label_map
        })
        return base_dict

    @classmethod
    def from_dict(cls, data, label_window):
        """Instantiate a MaskAnnotation from a dictionary."""
        # Get all labels currently in the project. This is needed for the constructor.
        all_project_labels = list(label_window.labels)
        if not all_project_labels:
            raise ValueError("Cannot import a MaskAnnotation without any labels loaded in the project.")

        # Decode the RLE mask data
        shape = tuple(data['shape'])
        mask_data = np.zeros(shape, dtype=np.uint8)
        for item in data['rle_masks']:
            class_id = item['class_id']
            rle = item['rle']
            try:
                rle['counts'] = base64.b64decode(rle['counts'])
                binary_mask = mask.decode(rle).astype(bool)
                if binary_mask.shape != shape:
                    print(f"Warning: RLE decoded shape {binary_mask.shape} does not match expected shape {shape}")
                    continue
                mask_data[binary_mask] = class_id
            except Exception as e:
                print(f"Error decoding RLE for class {class_id}: {e}")
                continue

        # Create the base annotation instance. It will have a generic label map initially.
        annotation = cls(
            image_path=data['image_path'],
            mask_data=mask_data,
            initial_labels=all_project_labels
        )
        
        # Clear the generic maps created by the constructor.
        annotation.class_id_to_label_map.clear()
        annotation.label_id_to_class_id_map.clear()
        
        max_id_found = 0
        if 'label_map' in data and data['label_map']:
            # Iterate through the saved map {class_id: short_code}
            for cid_str, short_code in data['label_map'].items():
                class_id = int(cid_str)
                label = label_window.get_label_by_short_code(short_code)
                
                if label:
                    # Rebuild the maps with the correct associations
                    annotation.class_id_to_label_map[class_id] = label
                    annotation.label_id_to_class_id_map[label.id] = class_id
                    if class_id > max_id_found:
                        max_id_found = class_id
                else:
                    print(f"Warning: Label with short code '{short_code}' not found in project during mask import.")

        # Ensure the next class ID is set correctly to avoid future conflicts.
        annotation.next_class_id = max_id_found + 1

        # Restore other annotation properties
        annotation.id = data.get('id', annotation.id)
        annotation.data = data.get('data', {})
        
        return annotation

    @classmethod
    def from_rasterio(cls, file_path: str, image_path: str, all_labels: list):
        """Creates a MaskAnnotation instance by loading data from a raster file."""
        with rasterio.open(file_path) as src:
            mask_data = src.read(1)
            return cls(
                image_path=image_path,
                mask_data=mask_data,
                initial_labels=all_labels,
                rasterio_src=src
            )

    # --- Compatibility Methods ---
    def get_perimeter(self):
        height, width = self.mask_data.shape
        return 2 * (width + height)
    
    def get_polygon(self):
        height, width = self.mask_data.shape
        return QPolygonF(QRectF(0, 0, width, height))

    def __repr__(self):
        return (f"MaskAnnotation(id={self.id}, image_path={self.image_path}, "
                f"shape={self.mask_data.shape})")


import warnings

from PyQt5.QtCore import Qt, QPointF, QRectF
from PyQt5.QtGui import QMouseEvent

from coralnet_toolbox.Tools.QtTool import Tool

warnings.filterwarnings("ignore", category=DeprecationWarning)


# ----------------------------------------------------------------------------------------------------------------------
# Classes
# ----------------------------------------------------------------------------------------------------------------------


class ZoomTool(Tool):
    def __init__(self, annotation_window):
        super().__init__(annotation_window)
        self.cursor = Qt.ArrowCursor
        self.min_zoom_factor = 1.0
        self.initial_zoom = 1.0

    def calculate_min_zoom(self):
        """Calculate minimum zoom factor to fit image in view"""
        scene = self.annotation_window.scene
        if not scene:
            return 1.0
            
        view_rect = self.annotation_window.viewport().rect()
        scene_rect = scene.sceneRect()
        
        # Calculate ratios to fit width and height
        width_ratio = view_rect.width() / scene_rect.width() if scene_rect.width() else 1.0
        height_ratio = view_rect.height() / scene_rect.height() if scene_rect.height() else 1.0
        
        # Store initial zoom when first calculated
        min_zoom = max(min(width_ratio, height_ratio), 0.1)
        if self.initial_zoom == 1.0:
            self.initial_zoom = min_zoom
            self.annotation_window.zoom_factor = min_zoom
            
        return min_zoom

    def reset_zoom(self):
        """Reset zoom state"""
        self.initial_zoom = 1.0
        self.min_zoom_factor = 1.0
        self.annotation_window.zoom_factor = 1.0
    
    def center_image_in_view(self):
        """Center the image in the viewport"""
        if not self.annotation_window.scene:
            return
            
        # Get the scene rect (image boundaries)
        scene_rect = self.annotation_window.scene.sceneRect()
        
        # Center the view on the scene rect
        self.annotation_window.centerOn(scene_rect.center())
        
        # Update view dimensions in status bar
        self.annotation_window.viewChanged.emit(
            *self.annotation_window.get_image_dimensions())

    def wheelEvent(self, event: QMouseEvent):
        """Handle zoom events with proper boundary constraints and centering"""
        if not self.annotation_window.active_image:
            return
            
        # Determine zoom direction
        if event.angleDelta().y() > 0:
            factor = 1.1  # Zoom in
        else:
            factor = 0.9  # Zoom out

        # Calculate new zoom level
        new_zoom = self.annotation_window.zoom_factor * factor
        min_zoom = self.calculate_min_zoom()
        
        # Prevent zooming out beyond minimum (image boundaries)
        if new_zoom < min_zoom and factor < 1:
            # Instead of returning, set to minimum zoom and force centering
            new_zoom = min_zoom
            factor = new_zoom / self.annotation_window.zoom_factor
            
            # Apply zoom and save the new factor
            self.annotation_window.scale(factor, factor)
            self.annotation_window.zoom_factor = new_zoom
            
            # Center the image in view
            self.center_image_in_view()
            return
            
        # For normal zooming, store position before zoom
        old_pos = self.annotation_window.mapToScene(event.pos())
        
        # Apply zoom and save the new factor
        self.annotation_window.scale(factor, factor)
        self.annotation_window.zoom_factor = new_zoom
        
        # Correct position after zoom for a more natural zoom effect
        new_pos = self.annotation_window.mapToScene(event.pos())
        delta = new_pos - old_pos
        self.annotation_window.translate(delta.x(), delta.y())
        
        # When zoomed out completely, ensure perfect centering
        if abs(new_zoom - min_zoom) < 0.01:
            self.center_image_in_view()


# Credit: https://www.a1k0n.net/code/tron.html

import random
import warnings

warnings.filterwarnings("ignore")

from PyQt5.QtCore import Qt, QTimer
from PyQt5.QtGui import QPainter, QColor, QPen, QBrush, QFont
from PyQt5.QtWidgets import (QMainWindow, QMessageBox, QDialog, QVBoxLayout,
                             QHBoxLayout, QPushButton, QLabel, QButtonGroup)

from coralnet_toolbox.Icons import get_icon


# ----------------------------------------------------------------------------------------------------------------------
# Constants / Configurations
# ----------------------------------------------------------------------------------------------------------------------


# Game settings
BOARD_WIDTH = 80
BOARD_HEIGHT = 60
CELL_SIZE = 12
GRID_LINE_WIDTH = 2

# Colors
BACKGROUND_COLOR = QColor(0, 0, 0)  # Black
GRID_COLOR = QColor(0, 100, 100)  # Darker cyan
PLAYER_COLOR = QColor(255, 215, 0)  # Gold
OPPONENT_COLOR = QColor(255, 0, 0)  # Red

# Directions
UP = 0
DOWN = 1
LEFT = 2
RIGHT = 3

# Game states
PLAYER_ALIVE = 0
PLAYER_CRASHED = 1
OPPONENT_ALIVE = 0
OPPONENT_CRASHED = 1


# ----------------------------------------------------------------------------------------------------------------------
# Classes
# ----------------------------------------------------------------------------------------------------------------------


class DifficultyDialog(QDialog):
    """
    Dialog for selecting game difficulty level.
    """
    def __init__(self, parent=None):
        super().__init__(parent)
        self.selected_difficulty = "Medium"  # Default
        self.init_ui()

    def init_ui(self):
        """Initialize the difficulty selection UI."""
        self.setWindowTitle("Select Difficulty")
        self.setModal(True)
        self.setFixedSize(400, 300)

        layout = QVBoxLayout()

        # Title
        title = QLabel("Choose Difficulty Level")
        title.setAlignment(Qt.AlignCenter)
        title.setFont(title.font())
        title.font().setPointSize(16)
        title.font().setBold(True)
        layout.addWidget(title)

        # Difficulty descriptions
        difficulties = {
            "Easy": "🟢 Slower opponent, larger board\nPerfect for beginners!",
            "Medium": "🟡 Balanced gameplay\nThe classic experience",
            "Hard": "🟠 Faster opponent, smaller board\nFor experienced players",
            "Insane": "🔴 Lightning fast opponent\nOnly for the brave!"
        }

        self.button_group = QButtonGroup()

        for i, (difficulty, description) in enumerate(difficulties.items()):
            btn = QPushButton(f"{difficulty}\n{description}")
            btn.setFixedHeight(50)
            btn.setCheckable(True)
            btn.clicked.connect(lambda checked, d=difficulty: self.set_difficulty(d))
            self.button_group.addButton(btn, i)
            layout.addWidget(btn)

            if difficulty == "Medium":  # Default selection
                btn.setChecked(True)

        # OK/Cancel buttons
        button_layout = QHBoxLayout()
        ok_btn = QPushButton("Start Game")
        cancel_btn = QPushButton("Cancel")

        ok_btn.clicked.connect(self.accept)
        cancel_btn.clicked.connect(self.reject)

        button_layout.addWidget(cancel_btn)
        button_layout.addWidget(ok_btn)
        layout.addLayout(button_layout)

        self.setLayout(layout)

    def set_difficulty(self, difficulty):
        """Set the selected difficulty."""
        self.selected_difficulty = difficulty

    def get_difficulty(self):
        """Return the selected difficulty."""
        return self.selected_difficulty


class LightCycle:
    """
    Represents a light cycle in the game.
    """

    def __init__(self, start_x, start_y, direction, color):
        self.x = start_x
        self.y = start_y
        self.direction = direction
        self.color = color
        self.trail = [(start_x, start_y)]  # List of (x, y) positions
        self.alive = True

    def move(self):
        """Move the cycle in its current direction."""
        if not self.alive:
            return

        if self.direction == UP:
            self.y -= 1
        elif self.direction == DOWN:
            self.y += 1
        elif self.direction == LEFT:
            self.x -= 1
        elif self.direction == RIGHT:
            self.x += 1

        # Add current position to trail
        self.trail.append((self.x, self.y))

    def turn(self, new_direction):
        """Change direction if it's not reversing."""
        # Prevent immediate reversal
        if (self.direction == UP and new_direction == DOWN) or \
           (self.direction == DOWN and new_direction == UP) or \
           (self.direction == LEFT and new_direction == RIGHT) or \
           (self.direction == RIGHT and new_direction == LEFT):
            return
        self.direction = new_direction

    def check_collision(self, board_width, board_height, opponent_trail, obstacles):
        """Check if the cycle has crashed."""
        # Wall collision
        if self.x < 0 or self.x >= board_width or self.y < 0 or self.y >= board_height:
            self.alive = False
            return True

        # Self collision (hit own trail)
        if (self.x, self.y) in self.trail[:-1]:  # Exclude current position
            self.alive = False
            return True

        # Opponent collision
        if (self.x, self.y) in opponent_trail:
            self.alive = False
            return True

        # Obstacle collision
        if (self.x, self.y) in obstacles:
            self.alive = False
            return True

        return False


class LightCycleAI:
    """
    Advanced AI for Light Cycle based on Google AI Challenge winning strategies.
    Implements Voronoi heuristic, articulation points, space-filling, and minimax.
    """
    
    def __init__(self, difficulty="Medium"):
        self.difficulty = difficulty
        self.think_ahead = 4
        self.mistake_chance = 0.04
        
        # Difficulty settings
        difficulty_settings = {
            "Easy": {"think_ahead": 2, "mistake_chance": 0.15, "search_depth": 3},
            "Medium": {"think_ahead": 3, "mistake_chance": 0.08, "search_depth": 4},
            "Hard": {"think_ahead": 4, "mistake_chance": 0.04, "search_depth": 6},
            "Insane": {"think_ahead": 5, "mistake_chance": 0.01, "search_depth": 8}
        }
        
        if difficulty in difficulty_settings:
            settings = difficulty_settings[difficulty]
            self.think_ahead = settings["think_ahead"]
            self.mistake_chance = settings["mistake_chance"]
            self.search_depth = settings["search_depth"]
        else:
            self.search_depth = 5
            
        # Voronoi heuristic coefficients (from data mining in original contest)
        self.node_coefficient = 0.080
        self.edge_coefficient = 0.150

    def compute_voronoi_territories(self, board_width, board_height, occupied_positions, player_pos, opponent_pos):
        """
        Compute Voronoi diagram - for each cell, determine which player can reach it first.
        Returns (player_territory, opponent_territory, player_edges, opponent_edges)
        """
        # Use BFS to find shortest distance from each player to all reachable cells
        distances = {}
        territories = {}
        
        # Initialize BFS for both players
        from collections import deque
        queue = deque()
        
        # Add starting positions
        if player_pos not in occupied_positions:
            queue.append((player_pos[0], player_pos[1], 0, 'player'))
            distances[player_pos] = 0
            territories[player_pos] = 'player'
            
        if opponent_pos not in occupied_positions:
            queue.append((opponent_pos[0], opponent_pos[1], 0, 'opponent'))
            distances[opponent_pos] = 0
            territories[opponent_pos] = 'opponent'
        
        # BFS to find territories
        while queue:
            x, y, dist, owner = queue.popleft()
            current_pos = (x, y)
            
            # Skip if we've found a shorter path already
            if current_pos in distances and distances[current_pos] < dist:
                continue
                
            # Explore neighbors
            for dx, dy in [(0, 1), (0, -1), (1, 0), (-1, 0)]:
                nx, ny = x + dx, y + dy
                new_pos = (nx, ny)
                
                # Check bounds and walls
                if (nx < 0 or 
                    nx >= board_width or 
                    ny < 0 or ny >= board_height or 
                    new_pos in occupied_positions):
                    continue
                    
                new_dist = dist + 1
                
                # If unvisited or we found a shorter path
                if new_pos not in distances or distances[new_pos] > new_dist:
                    distances[new_pos] = new_dist
                    territories[new_pos] = owner
                    queue.append((nx, ny, new_dist, owner))
                elif distances[new_pos] == new_dist and territories[new_pos] != owner:
                    # Tie - mark as neutral
                    territories[new_pos] = 'neutral'
        
        # Count territories and edges
        player_nodes = sum(1 for pos, owner in territories.items() if owner == 'player')
        opponent_nodes = sum(1 for pos, owner in territories.items() if owner == 'opponent')
        
        # Count edges (neighbors of territory cells)
        player_edges = 0
        opponent_edges = 0
        
        for (x, y), owner in territories.items():
            if owner == 'player':
                # Count open neighbors
                for dx, dy in [(0, 1), (0, -1), (1, 0), (-1, 0)]:
                    nx, ny = x + dx, y + dy
                    if (0 <= nx < board_width and 
                        0 <= ny < board_height and
                        (nx, ny) not in occupied_positions):
                        player_edges += 1
            elif owner == 'opponent':
                for dx, dy in [(0, 1), (0, -1), (1, 0), (-1, 0)]:
                    nx, ny = x + dx, y + dy
                    if (0 <= nx < board_width and 
                        0 <= ny < board_height and
                        (nx, ny) not in occupied_positions):
                        opponent_edges += 1
        
        # Find neutral (battlefront) squares
        neutral_squares = {pos for pos, owner in territories.items() if owner == 'neutral'}
        
        return player_nodes, opponent_nodes, player_edges, opponent_edges, neutral_squares

    def find_articulation_points(self, board_width, board_height, occupied_positions):
        """
        Find articulation points (cut vertices) in the graph of free spaces.
        These are points whose removal would disconnect the graph.
        """
        # Build adjacency list for free spaces
        free_spaces = []
        space_to_index = {}
        
        for x in range(board_width):
            for y in range(board_height):
                if (x, y) not in occupied_positions:
                    space_to_index[(x, y)] = len(free_spaces)
                    free_spaces.append((x, y))
        
        if len(free_spaces) <= 2:
            return set()
            
        n = len(free_spaces)
        adj = [[] for _ in range(n)]
        
        # Build adjacency list
        for i, (x, y) in enumerate(free_spaces):
            for dx, dy in [(0, 1), (0, -1), (1, 0), (-1, 0)]:
                nx, ny = x + dx, y + dy
                if (nx, ny) in space_to_index:
                    j = space_to_index[(nx, ny)]
                    adj[i].append(j)
        
        # Find articulation points using Tarjan's algorithm
        visited = [False] * n
        disc = [0] * n
        low = [0] * n
        parent = [-1] * n
        ap = [False] * n
        time = [0]
        
        def bridge_util(u):
            children = 0
            visited[u] = True
            disc[u] = low[u] = time[0]
            time[0] += 1
            
            for v in adj[u]:
                if not visited[v]:
                    children += 1
                    parent[v] = u
                    bridge_util(v)
                    
                    low[u] = min(low[u], low[v])
                    
                    # Root of DFS tree is articulation point if it has more than one child
                    if parent[u] == -1 and children > 1:
                        ap[u] = True
                        
                    # Non-root is articulation point if removing it disconnects the tree
                    if parent[u] != -1 and low[v] >= disc[u]:
                        ap[u] = True
                        
                elif v != parent[u]:
                    low[u] = min(low[u], disc[v])
        
        # Find articulation points for all components
        for i in range(n):
            if not visited[i]:
                bridge_util(i)
        
        # Convert back to coordinates
        articulation_points = set()
        for i in range(n):
            if ap[i]:
                articulation_points.add(free_spaces[i])
                
        return articulation_points

    def compute_checkerboard_territories(self, board_width, board_height, occupied_positions, player_pos, opponent_pos):
        """
        Compute territories considering checkerboard parity constraints.
        In Tron, players alternate between "red" and "black" squares, so surplus squares
        of one color will be wasted.
        """
        player_nodes, opponent_nodes, player_edges, opponent_edges, _ = self.compute_voronoi_territories(
            board_width, board_height, occupied_positions, player_pos, opponent_pos)
        
        # Count red/black squares for each player's territory
        player_red = 0
        player_black = 0
        opponent_red = 0
        opponent_black = 0
        
        # Re-traverse territories to count colors
        distances = {}
        territories = {}
        
        # BFS again to get territories
        from collections import deque
        queue = deque()
        
        if player_pos not in occupied_positions:
            queue.append((player_pos[0], player_pos[1], 0, 'player'))
            distances[player_pos] = 0
            territories[player_pos] = 'player'
            
        if opponent_pos not in occupied_positions:
            queue.append((opponent_pos[0], opponent_pos[1], 0, 'opponent'))
            distances[opponent_pos] = 0
            territories[opponent_pos] = 'opponent'
        
        while queue:
            x, y, dist, owner = queue.popleft()
            current_pos = (x, y)
            
            if current_pos in distances and distances[current_pos] < dist:
                continue
                
            for dx, dy in [(0, 1), (0, -1), (1, 0), (-1, 0)]:
                nx, ny = x + dx, y + dy
                new_pos = (nx, ny)
                
                if (nx < 0 or nx >= board_width or ny < 0 or ny >= board_height or
                    new_pos in occupied_positions):
                    continue
                    
                new_dist = dist + 1
                
                if new_pos not in distances or distances[new_pos] > new_dist:
                    distances[new_pos] = new_dist
                    territories[new_pos] = owner
                    queue.append((nx, ny, new_dist, owner))
                elif distances[new_pos] == new_dist and territories[new_pos] != owner:
                    territories[new_pos] = 'neutral'
        
        # Count red/black squares
        for (x, y), owner in territories.items():
            is_red = (x + y) % 2 == 0  # Checkerboard pattern
            if owner == 'player':
                if is_red:
                    player_red += 1
                else:
                    player_black += 1
            elif owner == 'opponent':
                if is_red:
                    opponent_red += 1
                else:
                    opponent_black += 1
        
        # Calculate effective territory considering parity constraints
        player_effective = min(player_red, player_black) * 2 + abs(player_red - player_black)
        opponent_effective = min(opponent_red, opponent_black) * 2 + abs(opponent_red - opponent_black)
        
        return player_effective, opponent_effective

    def count_edge_removal(self, board_width, board_height, occupied_positions, x, y):
        """
        Count how many edges would be removed by moving to position (x, y).
        Lower values indicate better space-filling moves.
        """
        if (x, y) in occupied_positions:
            return float('inf')
            
        edge_count = 0
        for dx, dy in [(0, 1), (0, -1), (1, 0), (-1, 0)]:
            nx, ny = x + dx, y + dy
            if (0 <= nx < board_width and 0 <= ny < board_height and
                (nx, ny) not in occupied_positions):
                edge_count += 1
                
        return 4 - edge_count  # Higher removal = worse move

    def evaluate_position(self, board_width, board_height, occupied_positions, player_pos, opponent_pos):
        """
        Advanced position evaluation using Voronoi heuristic and space-filling principles.
        Returns a score from the opponent's perspective (higher = better for opponent).
        """
        # Check for immediate game over
        if player_pos in occupied_positions:
            return 10000  # Opponent wins
        if opponent_pos in occupied_positions:
            return -10000  # Player wins
        
        # Compute Voronoi territories
        player_nodes, opponent_nodes, player_edges, opponent_edges, neutral_squares = self.compute_voronoi_territories(
            board_width, board_height, occupied_positions, player_pos, opponent_pos)
        
        # If players are separated, use territory count with endgame bonus
        if self.are_players_separated(board_width, board_height, occupied_positions, player_pos, opponent_pos):
            # Use checkerboard-aware territory calculation for endgame
            player_effective, opponent_effective = self.compute_checkerboard_territories(
                board_width, board_height, occupied_positions, player_pos, opponent_pos)
            territory_diff = opponent_effective - player_effective
            return territory_diff * 1000  # Large multiplier for certain endgame
        
        # Use learned coefficients from Google AI Challenge data mining
        node_score = (opponent_nodes - player_nodes) * self.node_coefficient
        edge_score = (opponent_edges - player_edges) * self.edge_coefficient
        
        base_score = (node_score + edge_score) * 1000  # Scale up for integer math
        
        # --- START OF AGGRESSIVENESS CODE ---
        # 1. Proximity Bonus: Reward AI for being close to the player.
        # The original formula (e.g., 5000 / distance) creates a weak gradient at long
        # range, leading to passive play. A linear bonus provides a stronger, constant
        # incentive to engage the opponent.
        distance = abs(player_pos[0] - opponent_pos[0]) + abs(player_pos[1] - opponent_pos[1])
        max_dist = board_width + board_height
        # A high weight (e.g., 500) makes the score change from moving one step closer
        # comparable to the score from gaining territory, encouraging aggression.
        proximity_bonus = (max_dist - distance) * 1000

        # 2. Battlefront Bonus: Reward AI for controlling the contested border.
        # Being on the "battlefront" (neutral squares) is a key strategic goal.
        # This bonus should be large to heavily incentivize moving to contested areas.
        battlefront_bonus = 0
        if opponent_pos in neutral_squares:
            battlefront_bonus = 10000
        # --- END OF AGGRESSIVENESS CODE ---
        
        # Add space-filling bonus - prefer moves that don't waste edges
        space_filling_bonus = 0
        
        # Find articulation points to avoid creating bottlenecks
        articulation_points = self.find_articulation_points(board_width, board_height, occupied_positions)
        
        # Penalty for being near articulation points (might get cut off)
        articulation_penalty = 0
        for ax, ay in articulation_points:
            opponent_dist = abs(opponent_pos[0] - ax) + abs(opponent_pos[1] - ay)
            player_dist = abs(player_pos[0] - ax) + abs(player_pos[1] - ay)
            if opponent_dist <= 2:
                articulation_penalty -= 50
            if player_dist <= 2:
                articulation_penalty += 50
        
        return int(base_score + space_filling_bonus + articulation_penalty + proximity_bonus + battlefront_bonus)

    def are_players_separated(self, board_width, board_height, occupied_positions, player_pos, opponent_pos):
        """
        Check if players are in separate connected components (endgame condition).
        """
        # BFS from player position
        visited = set()
        queue = [player_pos]
        visited.add(player_pos)
        
        while queue:
            x, y = queue.pop(0)
            
            if (x, y) == opponent_pos:
                return False  # Players are connected
                
            for dx, dy in [(0, 1), (0, -1), (1, 0), (-1, 0)]:
                nx, ny = x + dx, y + dy
                if (0 <= nx < board_width and 0 <= ny < board_height and
                    (nx, ny) not in occupied_positions and (nx, ny) not in visited):
                    visited.add((nx, ny))
                    queue.append((nx, ny))
        
        return True  # Opponent not reachable from player

    def minimax(self, board_width, board_height, occupied_positions, player_pos, player_dir, 
                opponent_pos, opponent_dir, depth, maximizing_player, alpha=float('-inf'), beta=float('inf')):
        """
        Minimax search with alpha-beta pruning.
        """
        # Base case - evaluate position
        if depth == 0:
            return self.evaluate_position(board_width, board_height, occupied_positions, player_pos, opponent_pos)
        
        if maximizing_player:  # Opponent's turn (AI)
            max_eval = float('-inf')
            for direction in [UP, DOWN, LEFT, RIGHT]:
                # Skip invalid moves (reversing direction)
                if ((opponent_dir == UP and direction == DOWN) or
                    (opponent_dir == DOWN and direction == UP) or
                    (opponent_dir == LEFT and direction == RIGHT) or
                    (opponent_dir == RIGHT and direction == LEFT)):
                    continue
                
                # Simulate move
                new_opponent_pos = self.simulate_move(opponent_pos, direction)
                if self.is_valid_move(board_width, board_height, occupied_positions, new_opponent_pos):
                    
                    # --- START OF PURSUIT REVERSAL BONUS ---
                    # This tactical bonus encourages the AI to make sharp turns
                    # to cut off a player it is trailing ("pursuing").
                    reversal_bonus = 0
                    distance = abs(player_pos[0] - opponent_pos[0]) + abs(player_pos[1] - opponent_pos[1])
                    if 2 < distance < 15:  # Only applies at medium-close range
                        is_behind = False
                        # Check if AI is generally "behind" the player, relative to its direction of travel.
                        if opponent_dir == RIGHT and player_pos[0] > opponent_pos[0]: 
                            is_behind = True
                        elif opponent_dir == LEFT and player_pos[0] < opponent_pos[0]: 
                            is_behind = True
                        elif opponent_dir == DOWN and player_pos[1] > opponent_pos[1]: 
                            is_behind = True
                        elif opponent_dir == UP and player_pos[1] < opponent_pos[1]: 
                            is_behind = True

                        is_turn = (opponent_dir != direction)

                        if is_behind and is_turn:
                            # A large bonus makes this a high-priority tactical move.
                            reversal_bonus = 8000
                    # --- END OF PURSUIT REVERSAL BONUS ---
                    
                    new_occupied = occupied_positions.union({new_opponent_pos})
                    eval_score = self.minimax(board_width, 
                                              board_height, 
                                              new_occupied, 
                                              player_pos, 
                                              player_dir,
                                              new_opponent_pos, 
                                              direction, 
                                              depth - 1, 
                                              False,
                                              alpha, 
                                              beta)
                    
                    eval_score += reversal_bonus  # Add bonus to the evaluated score for this move
                    
                    max_eval = max(max_eval, eval_score)
                    alpha = max(alpha, eval_score)
                    if beta <= alpha:
                        break  # Alpha-beta pruning
                else:
                    # Invalid move = loss
                    max_eval = max(max_eval, -10000)
            return max_eval
        else:  # Player's turn
            min_eval = float('inf')
            for direction in [UP, DOWN, LEFT, RIGHT]:
                # Skip invalid moves
                if ((player_dir == UP and direction == DOWN) or
                    (player_dir == DOWN and direction == UP) or
                    (player_dir == LEFT and direction == RIGHT) or
                    (player_dir == RIGHT and direction == LEFT)):
                    continue
                
                # Simulate move
                new_player_pos = self.simulate_move(player_pos, direction)
                if self.is_valid_move(board_width, board_height, occupied_positions, new_player_pos):
                    new_occupied = occupied_positions.union({new_player_pos})
                    eval_score = self.minimax(board_width, 
                                              board_height, 
                                              new_occupied, 
                                              new_player_pos, 
                                              direction,
                                              opponent_pos, 
                                              opponent_dir, 
                                              depth - 1, 
                                              True, 
                                              alpha, 
                                              beta)
                    
                    min_eval = min(min_eval, eval_score)
                    beta = min(beta, eval_score)
                    if beta <= alpha:
                        break  # Alpha-beta pruning
                else:
                    # Invalid move = opponent wins
                    min_eval = min(min_eval, 10000)
            return min_eval

    def simulate_move(self, pos, direction):
        """Simulate a move in the given direction."""
        x, y = pos
        if direction == UP:
            return (x, y - 1)
        elif direction == DOWN:
            return (x, y + 1)
        elif direction == LEFT:
            return (x - 1, y)
        elif direction == RIGHT:
            return (x + 1, y)
        return pos

    def is_valid_move(self, board_width, board_height, occupied_positions, pos):
        """Check if a move to the given position is valid."""
        x, y = pos
        return (0 <= x < board_width and 0 <= y < board_height and pos not in occupied_positions)

    def choose_direction(self, opponent, player, board_width, board_height, occupied_positions):
        """
        Choose the best direction using advanced AI techniques.
        """
        # Occasionally make mistakes based on difficulty
        if random.random() < self.mistake_chance:
            valid_directions = []
            for direction in [UP, DOWN, LEFT, RIGHT]:
                # Don't reverse direction
                if ((opponent.direction == UP and direction == DOWN) or
                    (opponent.direction == DOWN and direction == UP) or
                    (opponent.direction == LEFT and direction == RIGHT) or
                    (opponent.direction == RIGHT and direction == LEFT)):
                    continue
                    
                # Check if move is valid
                new_pos = self.simulate_move((opponent.x, opponent.y), direction)
                if self.is_valid_move(board_width, board_height, occupied_positions, new_pos):
                    valid_directions.append(direction)
            
            if valid_directions:
                return random.choice(valid_directions)
        
        # Use minimax search for best move
        best_direction = opponent.direction
        best_score = float('-inf')
        
        for direction in [UP, DOWN, LEFT, RIGHT]:
            # Don't reverse direction
            if ((opponent.direction == UP and direction == DOWN) or
                (opponent.direction == DOWN and direction == UP) or
                (opponent.direction == LEFT and direction == RIGHT) or
                (opponent.direction == RIGHT and direction == LEFT)):
                continue
            
            # Check if move is valid
            new_opponent_pos = self.simulate_move((opponent.x, opponent.y), direction)
            if not self.is_valid_move(board_width, board_height, occupied_positions, new_opponent_pos):
                continue
            
            # Use minimax to evaluate this move
            new_occupied = occupied_positions.union({new_opponent_pos})
            score = self.minimax(board_width, 
                                 board_height, 
                                 new_occupied,
                                 (player.x, player.y), 
                                 player.direction,
                                 new_opponent_pos, 
                                 direction, 
                                 min(self.search_depth, 4), 
                                 False)  # Start with player's turn
            
            if score > best_score:
                best_score = score
                best_direction = direction
        
        # Fallback to space-filling heuristic if no good moves found
        if best_score == float('-inf'):
            best_edge_removal = float('inf')
            for direction in [UP, DOWN, LEFT, RIGHT]:
                if ((opponent.direction == UP and direction == DOWN) or
                    (opponent.direction == DOWN and direction == UP) or
                    (opponent.direction == LEFT and direction == RIGHT) or
                    (opponent.direction == RIGHT and direction == LEFT)):
                    continue
                
                new_pos = self.simulate_move((opponent.x, opponent.y), direction)
                if self.is_valid_move(board_width, board_height, occupied_positions, new_pos):
                    edge_removal = self.count_edge_removal(board_width, 
                                                           board_height, 
                                                           occupied_positions,
                                                           new_pos[0], 
                                                           new_pos[1])
                    if edge_removal < best_edge_removal:
                        best_edge_removal = edge_removal
                        best_direction = direction
        
        return best_direction
    

class LightCycleGame(QMainWindow):
    """
    Main game window for the Light Cycle game.
    """

    def __init__(self, main_window, parent=None):
        super().__init__()
        self.main_window = main_window

        # Remove minimize and maximize buttons
        self.setWindowFlags(Qt.Window | Qt.CustomizeWindowHint |
                            Qt.WindowTitleHint | Qt.WindowCloseButtonHint)

        # Set the window icon
        self.setWindowIcon(get_icon("lightcycle.png"))
        # Set the window title
        self.title = "Light Cycle Game"

        # Game settings
        self.board_width = BOARD_WIDTH
        self.board_height = BOARD_HEIGHT
        self.cell_size = CELL_SIZE
        self.game_speed = 150  # milliseconds

        # Initialize timer
        self.timer = QTimer(self)
        self.difficulty = "Medium"

        # Initialize game
        self.player = None
        self.opponents = []  # List of AI opponents
        self.ais = []  # List of AI instances
        self.game_started = False
        self.title_timer = 0  # For title animation
        self.obstacles = set()  # Set of (x, y) positions for obstacles

    def start_game(self):
        """Start the game by initializing."""
        # Hardcode difficulty to Hard for AI opponent
        self.difficulty = "Hard"
        
        # Always reset game speed to baseline for this difficulty
        speed_settings = {
            "Easy": 200,
            "Medium": 150,
            "Hard": 100,
            "Insane": 75
        }
        self.game_speed = speed_settings[self.difficulty]
        
        # Stop any existing timer to prevent multiple timers running
        if hasattr(self, 'timer') and self.timer.isActive():
            self.timer.stop()

        # Show welcome dialog
        welcome_msg = (
            f"Welcome to Light Cycle - {self.difficulty} Mode!\n\n"
            "Rules:\n"
            " - Use WASD keys to turn your cycle (gold)\n"
            " - Avoid walls, your trail, and the opponents' trails (red)\n"
            " - The opponents move automatically\n"
            " - Last cycle standing wins!\n\n"
            "Click 'OK' to start playing."
        )
        if QMessageBox.information(self, "Welcome to Light Cycle", welcome_msg, QMessageBox.Ok) == QMessageBox.Ok:
            self.init_game()
            self.init_ui()
            # Disconnect any existing connection to avoid multiple calls
            try:
                self.timer.timeout.disconnect(self.update_game)
            except TypeError:
                pass  # Signal was not connected, which is fine
            self.timer.timeout.connect(self.update_game)
            self.timer.start(self.game_speed)
        else:
            self.close()

    def init_ui(self):
        """Set up the user interface."""
        window_width = self.board_width * self.cell_size
        window_height = self.board_height * self.cell_size

        self.resize(window_width, window_height)
        self.setFixedSize(window_width, window_height)
        self.setWindowTitle(self.title)
        self.show()

    def init_game(self):
        """Initialize the game state."""
        # Reset game state completely
        self.game_started = False
        
        # Player starts at bottom left, moving right
        self.player = LightCycle(5, self.board_height - 5, RIGHT, PLAYER_COLOR)

        # Create 5 opponents at different starting positions
        opponent_starts = [
            (self.board_width // 2, self.board_height // 2, DOWN),  # center
            (10, 10, DOWN),  # top-left
            (self.board_width - 10, 10, LEFT),  # top-right
            (10, self.board_height - 10, RIGHT),  # bottom-left
            (self.board_width - 10, self.board_height - 10, UP)  # bottom-right
        ]
        self.opponents = []
        for x, y, direction in opponent_starts:
            opponent = LightCycle(x, y, direction, OPPONENT_COLOR)
            self.opponents.append(opponent)

        # Create AI instances for each opponent
        self.ais = [LightCycleAI(self.difficulty) for _ in range(5)]

        # Generate 100 random obstacles
        self.obstacles = set()
        opponent_positions = [(x, y) for x, y, _ in opponent_starts]
        for _ in range(100):
            while True:
                x = random.randint(0, self.board_width - 1)
                y = random.randint(0, self.board_height - 1)
                pos = (x, y)
                player_start = (5, self.board_height - 5)
                if pos not in [player_start] + opponent_positions and pos not in self.obstacles:
                    self.obstacles.add(pos)
                    break

        self.game_started = True
        self.title_timer = 10  # Show title for 0.5 seconds (30 frames at ~60fps)

    def paintEvent(self, event):
        """Render the game board."""
        painter = QPainter()
        painter.begin(self)

        # Draw background
        painter.fillRect(self.rect(), BACKGROUND_COLOR)

        # Draw grid with synth-wave glow
        grid_pen = QPen(GRID_COLOR, 2)  # Thicker lines
        painter.setPen(grid_pen)
        for x in range(0, self.board_width * self.cell_size, self.cell_size):
            painter.drawLine(x, 0, x, self.board_height * self.cell_size)
        for y in range(0, self.board_height * self.cell_size, self.cell_size):
            painter.drawLine(0, y, self.board_width * self.cell_size, y)

        # Draw trails with glow effect
        if self.player:
            # Draw glow/shadow first
            painter.setPen(Qt.NoPen)
            glow_color = QColor(255, 0, 255, 100)  # Semi-transparent magenta glow
            painter.setBrush(QBrush(glow_color))
            for x, y in self.player.trail:
                painter.drawRect((x * self.cell_size) - 1, (y * self.cell_size) - 1,
                                 self.cell_size + 2, self.cell_size + 2)

            # Draw main trail
            painter.setBrush(QBrush(self.player.color))
            for x, y in self.player.trail:
                painter.drawRect(x * self.cell_size, y * self.cell_size,
                                 self.cell_size, self.cell_size)

        for opponent in self.opponents:
            # Draw glow/shadow first
            painter.setPen(Qt.NoPen)
            glow_color = QColor(0, 255, 255, 100)  # Semi-transparent cyan glow
            painter.setBrush(QBrush(glow_color))
            for x, y in opponent.trail:
                painter.drawRect((x * self.cell_size) - 1, (y * self.cell_size) - 1,
                                 self.cell_size + 2, self.cell_size + 2)

            # Draw main trail
            painter.setBrush(QBrush(opponent.color))
            for x, y in opponent.trail:
                painter.drawRect(x * self.cell_size, y * self.cell_size,
                                 self.cell_size, self.cell_size)

        # Draw obstacles
        painter.setPen(Qt.NoPen)
        painter.setBrush(QBrush(QColor(128, 128, 128)))  # Gray
        for x, y in self.obstacles:
            painter.drawRect(x * self.cell_size, y * self.cell_size, self.cell_size, self.cell_size)

        painter.end()

        # Draw synth-wave title animation
        if self.title_timer > 0:
            painter.begin(self)
            title_font = QFont('Arial', 36, QFont.Bold)
            painter.setFont(title_font)

            # Create animated colors
            colors = [QColor(255, 0, 255), QColor(0, 255, 255), QColor(0, 255, 0), QColor(255, 255, 0)]
            color_index = (self.title_timer // 10) % len(colors)
            painter.setPen(colors[color_index])

            title_text = "LIGHT CYCLES"
            text_rect = painter.fontMetrics().boundingRect(title_text)
            x = (self.width() - text_rect.width()) // 2
            y = self.height() // 2

            # Add glow effect to title
            glow_colors = [QColor(255, 0, 255, 150), QColor(0, 255, 255, 150)]
            glow_index = (self.title_timer // 5) % len(glow_colors)
            painter.setPen(glow_colors[glow_index])
            painter.drawText(x - 2, y - 2, title_text)
            painter.drawText(x + 2, y - 2, title_text)
            painter.drawText(x - 2, y + 2, title_text)
            painter.drawText(x + 2, y + 2, title_text)

            painter.setPen(colors[color_index])
            painter.drawText(x, y, title_text)
            painter.end()

    def update_game(self):
        """Handle game updates."""
        # Safety check - don't update if game is not properly initialized
        if not hasattr(self, 'game_started') or not self.game_started:
            return
            
        if self.title_timer > 0:
            self.title_timer -= 1
            self.update()
            return

        if self.game_started:
            # Move player once per frame at constant speed
            self.player.move()
            
            # Calculate current occupied positions
            all_trails = self.player.trail + [pos for opp in self.opponents for pos in opp.trail]
            occupied_positions = set(all_trails) | self.obstacles
            
            # AI decides each opponent's next move
            intended_moves = {}
            for i, opponent in enumerate(self.opponents):
                if opponent.alive:
                    best_direction = self.ais[i].choose_direction(
                        opponent, self.player, self.board_width, self.board_height, occupied_positions)
                    new_pos = self.ais[i].simulate_move((opponent.x, opponent.y), best_direction)
                    intended_moves[i] = (best_direction, new_pos)
            
            # Check for conflicts in intended positions
            new_positions = [pos for _, pos in intended_moves.values()]
            position_counts = {}
            for pos in new_positions:
                position_counts[pos] = position_counts.get(pos, 0) + 1
            
            # Resolve conflicts by re-choosing for conflicting opponents
            for i in intended_moves:
                direction, new_pos = intended_moves[i]
                if position_counts[new_pos] > 1:
                    # Conflict, choose a random valid direction avoiding other intended moves
                    opponent = self.opponents[i]
                    valid_directions = []
                    for d in [UP, DOWN, LEFT, RIGHT]:
                        if ((opponent.direction == UP and d == DOWN) or
                            (opponent.direction == DOWN and d == UP) or
                            (opponent.direction == LEFT and d == RIGHT) or
                            (opponent.direction == RIGHT and d == LEFT)):
                            continue
                        np = self.ais[i].simulate_move((opponent.x, opponent.y), d)
                        # Occupied includes current + other intended (excluding this one's current intended)
                        other_intended = set([p for j, p in intended_moves.values() if j != i])
                        occ = occupied_positions | other_intended
                        if self.ais[i].is_valid_move(self.board_width, self.board_height, occ, np):
                            valid_directions.append(d)
                    if valid_directions:
                        new_dir = random.choice(valid_directions)
                        new_pos = self.ais[i].simulate_move((opponent.x, opponent.y), new_dir)
                        intended_moves[i] = (new_dir, new_pos)
            
            # Apply the moves
            for i, opponent in enumerate(self.opponents):
                if opponent.alive:
                    direction, _ = intended_moves[i]
                    opponent.turn(direction)
                    opponent.move()

            # Check player collision
            all_opponent_trails = [pos for opp in self.opponents for pos in opp.trail]
            player_crashed = self.player.check_collision(self.board_width, 
                                                         self.board_height, 
                                                         all_opponent_trails, 
                                                         self.obstacles)
            
            # Check each opponent collision
            for opponent in self.opponents:
                if opponent.alive:
                    other_trails = (self.player.trail +
                                    [pos for opp in self.opponents 
                                     if opp != opponent for pos in opp.trail])
                    opponent_crashed = opponent.check_collision(
                        self.board_width, self.board_height, other_trails, self.obstacles)

            # Check win conditions
            alive_opponents = sum(1 for opp in self.opponents if opp.alive)
            if player_crashed:
                self.game_over("You crashed! Opponents win.")
            elif alive_opponents == 0:
                self.game_over("All opponents crashed! You win!")

            self.update()

    def keyPressEvent(self, event):
        """Handle key presses for player control."""
        if not self.game_started or not self.player.alive:
            return

        key = event.key()
        
        # Handle movement
        if key == Qt.Key_W:
            self.player.turn(UP)
        elif key == Qt.Key_S:
            self.player.turn(DOWN)
        elif key == Qt.Key_A:
            self.player.turn(LEFT)
        elif key == Qt.Key_D:
            self.player.turn(RIGHT)

    def keyReleaseEvent(self, event):
        """Handle key releases (no longer needed for speed control)."""
        pass

    def game_over(self, message):
        """Handle game over."""
        # Stop the timer immediately to prevent multiple calls
        if hasattr(self, 'timer') and self.timer.isActive():
            self.timer.stop()
        
        # Reset game state to prevent further updates
        self.game_started = False
        
        # Show game over dialog
        reply = QMessageBox.question(self, "Game Over", f"{message}\n\nPlay again?",
                                     QMessageBox.Yes | QMessageBox.No)
        if reply == QMessageBox.Yes:
            self.start_game()
        else:
            # Close the window when user doesn't want to play again
            self.close()

    def closeEvent(self, event):
        """Handle window close."""
        # Ensure timer is stopped and game state is reset
        if hasattr(self, 'timer') and self.timer.isActive():
            self.timer.stop()
        self.game_started = False
        event.accept()

        # Disconnect any existing connection to avoid multiple calls
        try:
            self.timer.timeout.disconnect(self.update_game)
        except TypeError:
            pass  # Signal was not connected, which is fine

        self.timer.timeout.connect(self.update_game)



import warnings

import os
import uuid
import yaml
import glob
import shutil
import ujson as json

from PyQt5.QtCore import Qt, QPointF, QObject, QThread, pyqtSignal
from PyQt5.QtWidgets import (QFileDialog, QApplication, QMessageBox, QVBoxLayout, QGroupBox,
                             QLabel, QLineEdit, QDialog, QPushButton, QDialogButtonBox,
                             QGridLayout, QScrollArea, QFrame, QCheckBox, QRadioButton,
                             QToolButton)

from coralnet_toolbox.Annotations.QtPolygonAnnotation import PolygonAnnotation
from coralnet_toolbox.Annotations.QtRectangleAnnotation import RectangleAnnotation

from coralnet_toolbox.QtProgressBar import ProgressBar
from coralnet_toolbox.utilities import rasterio_open
from coralnet_toolbox.Icons import get_icon

warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.filterwarnings("ignore", category=UserWarning)


# ----------------------------------------------------------------------------------------------------------------------
# Worker Class for Threading
# ----------------------------------------------------------------------------------------------------------------------


class DatasetProcessor(QObject):
    """
    Worker object to process a dataset in a separate thread.
    It is completely decoupled from the GUI.
    """
    status_changed = pyqtSignal(str, int)
    progress_updated = pyqtSignal(int)
    processing_complete = pyqtSignal(list, list, list)
    error = pyqtSignal(str)
    finished = pyqtSignal()

    def __init__(self, yaml_path, output_folder, task, import_as, rename_on_conflict=False,
                 excluded_classes=None, image_import_policy='annotated_only', parent=None):
        super().__init__(parent)
        self.yaml_path = yaml_path
        self.output_folder = output_folder
        self.task = task  # 'detect' or 'segment' (source format)
        self.import_as = import_as  # 'rectangle' or 'polygon' (target format)
        self.rename_on_conflict = rename_on_conflict
        self.excluded_classes = excluded_classes if excluded_classes is not None else set()
        self.image_import_policy = image_import_policy
        self.is_running = True
        self.parsing_errors = []  # To collect errors instead of printing

    def stop(self):
        self.is_running = False

    # In class DatasetProcessor, update this method:

    def run(self):
        """Main processing method executed in the thread."""
        try:
            # Step 1: Read YAML and discover files
            with open(self.yaml_path, 'r') as file:
                data = yaml.safe_load(file)
            class_names = data.get('names', [])

            source_image_label_map = self._find_source_files()

            if not source_image_label_map:
                self.error.emit("No valid image/label pairs found in the dataset.")
                self.finished.emit()
                return

            # --- Step 2: Copy files with progress reporting ---
            self.status_changed.emit("Copying image files...", len(source_image_label_map))
            image_label_paths = self._copy_files_with_progress(source_image_label_map)

            if not self.is_running:
                self.finished.emit()
                return

            # Step 3: Parse label files and create annotation data
            self.status_changed.emit("Importing annotations...", len(image_label_paths))
            raw_annotations = self._create_raw_annotations(image_label_paths, class_names)

            if not self.is_running:
                self.finished.emit()
                return

            # Step 4 (REMOVED): The JSON export is no longer done here.

            # Step 5: Emit results for GUI to consume
            image_paths = list(image_label_paths.keys())
            self.processing_complete.emit(raw_annotations, image_paths, self.parsing_errors)

        except Exception as e:
            self.error.emit(f"An error occurred during processing: {str(e)}")
        finally:
            self.finished.emit()

    def _find_source_files(self):
        """Finds all source image and label paths based on the import policy."""
        dir_path = os.path.dirname(self.yaml_path)
        image_paths = glob.glob(f"{dir_path}/**/images/*.*", recursive=True)
        label_paths = glob.glob(f"{dir_path}/**/labels/*.txt", recursive=True)

        source_map = {}
        if self.image_import_policy == 'all':
            # Policy: Find all images, and match labels to them if they exist
            label_basenames_map = {os.path.splitext(os.path.basename(p))[0]: p for p in label_paths}
            for img_path in image_paths:
                img_basename = os.path.splitext(os.path.basename(img_path))[0]
                label_path = label_basenames_map.get(img_basename, None)  # Label can be None
                source_map[img_path] = label_path
        else:
            # Policy: Only find images that have a corresponding label file
            image_basenames_map = {os.path.splitext(os.path.basename(p))[0]: p for p in image_paths}
            for label_path in label_paths:
                label_basename_no_ext = os.path.splitext(os.path.basename(label_path))[0]
                if label_basename_no_ext in image_basenames_map:
                    src_image_path = image_basenames_map[label_basename_no_ext]
                    source_map[src_image_path] = label_path
        return source_map

    def _copy_files_with_progress(self, source_image_label_map):
        """Copies files and reports progress for each file."""
        img_out_dir = os.path.join(self.output_folder, "images")
        os.makedirs(img_out_dir, exist_ok=True)

        dest_label_map = {}
        for i, (src_image_path, label_path) in enumerate(source_image_label_map.items()):
            if not self.is_running:
                break

            original_img_basename = os.path.basename(src_image_path)

            if self.rename_on_conflict:
                base, ext = os.path.splitext(original_img_basename)
                unique_id = str(uuid.uuid4())[:8]
                new_img_basename = f"{base}_{unique_id}{ext}"
            else:
                new_img_basename = original_img_basename

            dest_image_path = os.path.join(img_out_dir, new_img_basename)
            shutil.copy(src_image_path, dest_image_path)

            dest_label_map[dest_image_path.replace("\\", "/")] = label_path
            self.progress_updated.emit(i + 1)

        return dest_label_map

    def _create_raw_annotations(self, image_label_paths, class_names):
        """
        Parses label files, converts format if needed, and creates raw annotation data.
        Returns a list of annotation dictionaries.
        """
        all_raw_annotations = []
        for i, (image_path, label_path) in enumerate(image_label_paths.items()):
            if not self.is_running:
                break

            # If there's no label file for this image (e.g., 'import all' policy), skip to progress update
            if not label_path:
                self.progress_updated.emit(i + 1)
                continue

            image_height, image_width = rasterio_open(image_path).shape
            with open(label_path, 'r') as file:
                lines = file.readlines()

            for line_num, line in enumerate(lines):
                try:
                    parts = list(map(float, line.split()))
                    class_id = int(parts[0])

                    class_name = class_names[class_id]
                    if class_name in self.excluded_classes:
                        continue  # Skip this annotation if its class was unchecked

                    raw_ann_data = {"image_path": image_path, "class_name": class_name}

                    parsed_data = {}
                    if self.task == 'detect':  # Source is bbox: class, x_c, y_c, w, h
                        _, x_c, y_c, w, h = parts
                        x = x_c * image_width
                        y = y_c * image_height
                        width = w * image_width
                        height = h * image_height
                        parsed_data['top_left'] = (x - width / 2, y - height / 2)
                        parsed_data['bottom_right'] = (x + width / 2, y + height / 2)
                    else:  # Source is polygon: class, x1, y1, x2, y2, ...
                        points_norm = parts[1:]
                        # Convert normalized coordinates to pixel coordinates
                        # Extract x and y coordinates from the flattened list
                        x_coords = points_norm[::2]  # Every even index (0, 2, 4...)
                        y_coords = points_norm[1::2]  # Every odd index (1, 3, 5...)
                        
                        # Scale coordinates by image dimensions
                        points = []
                        for x, y in zip(x_coords, y_coords):
                            pixel_x = x * image_width
                            pixel_y = y * image_height
                            points.append((pixel_x, pixel_y))
                        parsed_data['points'] = points

                    if self.import_as == 'rectangle':
                        raw_ann_data["type"] = "RectangleAnnotation"
                        if 'top_left' in parsed_data:
                            raw_ann_data.update(parsed_data)
                        else:
                            points = parsed_data['points']
                            x_coords = [p[0] for p in points]
                            y_coords = [p[1] for p in points]
                            raw_ann_data["top_left"] = (min(x_coords), min(y_coords))
                            raw_ann_data["bottom_right"] = (max(x_coords), max(y_coords))
                    elif self.import_as == 'polygon':
                        raw_ann_data["type"] = "PolygonAnnotation"
                        if 'points' in parsed_data:
                            raw_ann_data.update(parsed_data)
                        else:
                            tl, br = parsed_data['top_left'], parsed_data['bottom_right']
                            raw_ann_data["points"] = [(tl[0], tl[1]), (br[0], tl[1]), (br[0], br[1]), (tl[0], br[1])]
                    all_raw_annotations.append(raw_ann_data)
                except (ValueError, IndexError) as e:
                    error_msg = (f"In file '{os.path.basename(label_path)}' on line {line_num + 1}:\n"
                                 f"Skipped malformed content: '{line.strip()}'\nReason: {e}\n")
                    self.parsing_errors.append(error_msg)

            self.progress_updated.emit(i + 1)
        return all_raw_annotations


# ----------------------------------------------------------------------------------------------------------------------
# Dialog Classes
# ----------------------------------------------------------------------------------------------------------------------


class Base(QDialog):
    def __init__(self, main_window, parent=None):
        super(Base, self).__init__(parent)
        self.main_window = main_window
        self.annotation_window = main_window.annotation_window
        self.image_window = main_window.image_window

        self.setWindowIcon(get_icon("coral.png"))
        self.setWindowTitle("Import Dataset")
        self.resize(500, 350)

        self.task = None
        self.progress_bar = None
        self.thread = None
        self.worker = None
        self.output_folder = None
        self.class_checkboxes = []

        self.layout = QVBoxLayout(self)
        self.setup_info_layout()
        self.setup_yaml_layout()
        self.setup_output_layout()
        self.setup_buttons_layout()

        self.advanced_options_toggle.setEnabled(False)
        self.advanced_options_frame.setVisible(False)

    def setup_info_layout(self):
        raise NotImplementedError("Subclasses must implement method.")

    def setup_yaml_layout(self):
        """Set up the layout for selecting the data YAML file."""
        group_box = QGroupBox("Data YAML File")
        layout = QGridLayout()
        layout.addWidget(QLabel("File:"), 0, 0)
        self.yaml_path_label = QLineEdit()
        self.yaml_path_label.setReadOnly(True)
        self.yaml_path_label.setPlaceholderText("Select Data YAML file...")
        layout.addWidget(self.yaml_path_label, 0, 1)
        self.browse_yaml_button = QPushButton("Browse")
        self.browse_yaml_button.clicked.connect(self.browse_data_yaml)
        layout.addWidget(self.browse_yaml_button, 0, 2)
        group_box.setLayout(layout)
        self.layout.addWidget(group_box)

    def setup_output_layout(self):
        """Set up the layout for output directory and the advanced options accordion."""
        group_box = QGroupBox("Output Settings")
        layout = QGridLayout()
        layout.addWidget(QLabel("Directory:"), 0, 0)
        self.output_dir_label = QLineEdit()
        self.output_dir_label.setPlaceholderText("Select output directory...")
        layout.addWidget(self.output_dir_label, 0, 1)
        self.browse_output_button = QPushButton("Browse")
        self.browse_output_button.clicked.connect(self.browse_output_dir)
        layout.addWidget(self.browse_output_button, 0, 2)
        layout.addWidget(QLabel("Folder Name:"), 1, 0)
        self.output_folder_name = QLineEdit("data")
        self.output_folder_name.setPlaceholderText("data")
        layout.addWidget(self.output_folder_name, 1, 1, 1, 2)
        group_box.setLayout(layout)
        self.layout.addWidget(group_box)

        self.advanced_options_toggle = QToolButton()
        self.advanced_options_toggle.setText("Advanced Options")
        self.advanced_options_toggle.setCheckable(True)
        self.advanced_options_toggle.setStyleSheet("QToolButton { border: none; }")
        self.advanced_options_toggle.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
        self.advanced_options_toggle.setArrowType(Qt.RightArrow)
        self.advanced_options_toggle.toggled.connect(self.toggle_advanced_options)
        self.layout.addWidget(self.advanced_options_toggle)

        self.advanced_options_frame = QFrame()
        self.advanced_options_frame.setFrameShape(QFrame.StyledPanel)
        advanced_layout = QVBoxLayout(self.advanced_options_frame)

        image_rule_box = QGroupBox("Image Import Rule")
        image_rule_layout = QVBoxLayout()
        self.import_annotated_images_radio = QRadioButton("Import only images with annotations")
        self.import_all_images_radio = QRadioButton("Import all images found in dataset")
        self.import_annotated_images_radio.setChecked(True)
        image_rule_layout.addWidget(self.import_annotated_images_radio)
        image_rule_layout.addWidget(self.import_all_images_radio)
        image_rule_box.setLayout(image_rule_layout)
        advanced_layout.addWidget(image_rule_box)

        class_filter_box = QGroupBox("Classes to Import")
        class_filter_layout = QVBoxLayout()
        self.class_scroll_area = QScrollArea()
        self.class_scroll_area.setWidgetResizable(True)
        self.class_widget = QFrame()
        self.class_layout = QVBoxLayout(self.class_widget)
        self.class_scroll_area.setWidget(self.class_widget)
        class_filter_layout.addWidget(self.class_scroll_area)
        class_filter_box.setLayout(class_filter_layout)
        advanced_layout.addWidget(class_filter_box)
        self.layout.addWidget(self.advanced_options_frame)

    def toggle_advanced_options(self, checked):
        self.advanced_options_toggle.setArrowType(Qt.DownArrow if checked else Qt.RightArrow)
        self.advanced_options_frame.setVisible(checked)

    def setup_buttons_layout(self):
        """Set up the OK/Cancel button box."""
        self.button_box = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        self.button_box.accepted.connect(self.start_processing)
        self.button_box.rejected.connect(self.reject)
        self.layout.addWidget(self.button_box)

    # In: QtBase.py

    def browse_data_yaml(self):
        """Open a file dialog to select the data YAML file and populate advanced options."""
        options = QFileDialog.Options()
        file_path, _ = QFileDialog.getOpenFileName(
            self, "Select Data YAML", "", "YAML Files (*.yaml);;All Files (*)", options=options
        )
        if not file_path:
            return

        try:
            with open(file_path, 'r') as file:
                data = yaml.safe_load(file)
            
            names_data = data.get('names')
            if not names_data:
                QMessageBox.warning(self, "Warning", "Could not find a 'names' entry in the selected YAML file.")
                return

            # Handle both dictionary and list formats for class names
            names_to_display = []
            if isinstance(names_data, dict):
                # If it's a dictionary (e.g., {0: 'coral'}), extract the values, sorting by key
                # to preserve the intended class order.
                names_to_display = [str(names_data[key]) for key in sorted(names_data.keys())]
            elif isinstance(names_data, list):
                # If it's already a list, use it directly.
                names_to_display = [str(name) for name in names_data]
            else:
                # Handle any other unexpected format.
                QMessageBox.warning(self, "Format Error", 
                                    f"The 'names' entry in the YAML has an unexpected format: {type(names_data)}.")
                return
            
            self.yaml_path_label.setText(file_path)
            yaml_dir = os.path.dirname(file_path)
            self.output_dir_label.setText(yaml_dir)
            self.output_folder_name.setText("data")

            # Clear any existing checkboxes before adding new ones
            for checkbox in self.class_checkboxes:
                self.class_layout.removeWidget(checkbox)
                checkbox.deleteLater()
            self.class_checkboxes.clear()

            # Create checkboxes using the processed list of names
            for name in names_to_display:
                checkbox = QCheckBox(name)
                checkbox.setChecked(True)
                self.class_layout.addWidget(checkbox)
                self.class_checkboxes.append(checkbox)

            self.advanced_options_toggle.setEnabled(True)

        except Exception as e:
            QMessageBox.critical(self, "Error", f"Failed to read or parse YAML file:\n{e}")
            self.advanced_options_toggle.setEnabled(False)

    def browse_output_dir(self):
        """Open a dialog to select the output directory."""
        dir_path = QFileDialog.getExistingDirectory(self, "Select Output Directory")
        if dir_path:
            self.output_dir_label.setText(dir_path)

    def start_processing(self):
        """Validate inputs, check for duplicates, and start the worker thread."""
        if not all([self.yaml_path_label.text(), self.output_dir_label.text(), self.output_folder_name.text()]):
            QMessageBox.warning(self, "Error", "Please fill in all fields.")
            return

        self.output_folder = os.path.join(self.output_dir_label.text(), self.output_folder_name.text())
        if os.path.exists(self.output_folder) and os.listdir(self.output_folder):
            reply = QMessageBox.question(self, 
                                         'Directory Not Empty', 
                                         f"The directory '{self.output_folder}' is not empty. Continue?", 
                                         QMessageBox.Yes | QMessageBox.No, QMessageBox.No)
            if reply == QMessageBox.No: return

        QApplication.setOverrideCursor(Qt.WaitCursor)
        try:
            image_paths = glob.glob(f"{os.path.dirname(self.yaml_path_label.text())}/**/images/*.*", recursive=True)
        finally:
            QApplication.restoreOverrideCursor()

        basenames, duplicates_exist = set(), False
        for path in image_paths:
            basename_no_ext = os.path.splitext(os.path.basename(path))[0]
            if basename_no_ext in basenames:
                duplicates_exist = True
                break
            basenames.add(basename_no_ext)

        rename_files = False
        if duplicates_exist:
            msg_box = QMessageBox(self)
            msg_box.setIcon(QMessageBox.Warning)
            msg_box.setWindowTitle('Duplicate Filenames Found')
            msg_box.setText(
                "Images with the same base name exist in different subdirectories.\n"
                "This can cause files to be overwritten in the output directory."
            )
            msg_box.setInformativeText("How would you like to handle these conflicts?")

            # Add buttons with proper line breaks
            rename_button = msg_box.addButton(
                "Rename Files (Safe)", 
                QMessageBox.AcceptRole
            )
            overwrite_button = msg_box.addButton(
                "Overwrite", 
                QMessageBox.DestructiveRole
            )
            cancel_button = msg_box.addButton(
                "Cancel", 
                QMessageBox.RejectRole
            )

            msg_box.setDefaultButton(rename_button)
            msg_box.exec_()

            clicked_button = msg_box.clickedButton()
            if clicked_button == cancel_button:
                return
            elif clicked_button == rename_button:
                rename_files = True
            elif clicked_button == overwrite_button:
                rename_files = False
            else:
                return

        excluded_classes = set()
        if self.advanced_options_toggle.isEnabled():
            for checkbox in self.class_checkboxes:
                if not checkbox.isChecked():
                    excluded_classes.add(checkbox.text())
        image_import_policy = 'all' if self.import_all_images_radio.isChecked() else 'annotated_only'

        self.button_box.setEnabled(False)
        QApplication.setOverrideCursor(Qt.WaitCursor)
        self.progress_bar = ProgressBar(self, title="Preparing to Import...")
        self.progress_bar.show()

        import_as = 'polygon' if 'Polygon' in self.import_as_combo.currentText() else 'rectangle'

        self.thread = QThread()
        self.worker = DatasetProcessor(
            yaml_path=self.yaml_path_label.text(),
            output_folder=self.output_folder,
            task=self.task,
            import_as=import_as,
            rename_on_conflict=rename_files,
            excluded_classes=excluded_classes,
            image_import_policy=image_import_policy
        )
        self.worker.moveToThread(self.thread)
        self.thread.started.connect(self.worker.run)
        self.worker.finished.connect(self.on_worker_finished)
        self.worker.error.connect(self.on_error)
        self.worker.status_changed.connect(self.on_status_changed)
        self.worker.progress_updated.connect(self.on_progress_update)
        self.worker.processing_complete.connect(self.on_processing_complete)
        self.thread.start()

    def on_status_changed(self, title, total):
        self.progress_bar.set_title(title)
        self.progress_bar.start_progress(total)

    def on_progress_update(self, value):
        self.progress_bar.set_value(value)

    def on_processing_complete(self, raw_annotations, image_paths, parsing_errors):
        progress_bar = ProgressBar(self, title="Adding Data to Project...")
        progress_bar.show()

        added_paths = []
        progress_bar.set_title(f"Adding {len(image_paths)} images...")
        progress_bar.start_progress(len(image_paths))
        
        for path in image_paths:
            # Call the manager directly to add the raster silently,
            # bypassing ImageWindow.add_image and its signal handlers.
            if self.image_window.raster_manager.add_raster(path, emit_signal=False):
                added_paths.append(path)
            progress_bar.update_progress()

        newly_created_annotations = []
        progress_bar.set_title(f"Adding {len(raw_annotations)} annotations...")
        progress_bar.start_progress(len(raw_annotations))
        for raw_ann in raw_annotations:
            label = self.main_window.label_window.add_label_if_not_exists(raw_ann["class_name"])
            if raw_ann["type"] == "RectangleAnnotation":
                tl, br = raw_ann["top_left"], raw_ann["bottom_right"]
                annotation = RectangleAnnotation(QPointF(tl[0], tl[1]), 
                                                 QPointF(br[0], br[1]), 
                                                 label.short_label_code, 
                                                 label.long_label_code, 
                                                 label.color, 
                                                 raw_ann["image_path"], 
                                                 label.id, 
                                                 self.main_window.get_transparency_value())
            else:
                points = [QPointF(p[0], p[1]) for p in raw_ann["points"]]
                annotation = PolygonAnnotation(points, 
                                               label.short_label_code, 
                                               label.long_label_code,
                                               label.color, 
                                               raw_ann["image_path"], 
                                               label.id, 
                                               self.main_window.get_transparency_value())

            self.annotation_window.add_annotation_to_dict(annotation)
            newly_created_annotations.append(annotation)
            
            progress_bar.update_progress()

        progress_bar.set_title("Exporting annotations.json...")
        self._export_annotations_to_json(newly_created_annotations, self.output_folder)
        
        progress_bar.finish_progress()
        progress_bar.stop_progress()
        progress_bar.close()
        
        # Manually perform a full UI update exactly once.
        self.image_window.update_search_bars()
        self.image_window.filter_images()

        if added_paths:
            self.image_window.load_image_by_path(added_paths[-1])
            self.image_window.update_image_annotations(added_paths[-1])
            self.annotation_window.load_annotations()

        summary_message = "Dataset has been successfully imported."
        if parsing_errors:
            QMessageBox.warning(self, 
                                "Import Complete with Warnings", 
                                f"{summary_message}\n\nHowever, {len(parsing_errors)} issue(s) were found. "
                                "Please review them below.", 
                                details='\n'.join(parsing_errors))
        else:
            QMessageBox.information(self, "Dataset Imported", summary_message)
            
    def _export_annotations_to_json(self, annotations_list, output_dir):
        """
        Merges the list of annotation objects into an existing annotations.json file,
        or creates a new one if it doesn't exist.
        The output is a dictionary mapping image paths to lists of annotation dicts.
        """
        export_dict = {}
        json_path = os.path.join(output_dir, "annotations.json")

        # Step 1: Check for the existing file and load it if present.
        if os.path.exists(json_path):
            try:
                with open(json_path, 'r') as file:
                    export_dict = json.load(file)
                # Ensure the loaded data is a dictionary
                if not isinstance(export_dict, dict):
                    raise TypeError("annotations.json is not in the expected format (dict).")
            except (json.JSONDecodeError, TypeError, IOError) as e:
                # If file is corrupt, unreadable, or has wrong format, warn the user and start fresh.
                QMessageBox.warning(self, 
                                    "Read Error",
                                    f"Could not read or parse existing annotations.json:\n{e}\n\n"
                                    "A new file will be created, overwriting the old one.")
                export_dict = {}  # Reset to be safe

        # Step 2: Iterate through new annotations and merge them into the dictionary.
        for annotation in annotations_list:
            image_path = annotation.image_path
            
            # Use setdefault to initialize a list for a new image path or get the existing one.
            export_dict.setdefault(image_path, [])
            
            # Create the dictionary for the annotation using its own method
            if isinstance(annotation, RectangleAnnotation):
                annotation_dict = {
                    'type': 'RectangleAnnotation',
                    **annotation.to_dict()
                }
            elif isinstance(annotation, PolygonAnnotation):
                annotation_dict = {
                    'type': 'PolygonAnnotation',
                    **annotation.to_dict()
                }
            else:
                warnings.warn(f"Unknown annotation type skipped during export: {type(annotation)}")
                continue

            export_dict[image_path].append(annotation_dict)

        # Step 3: Write the final, merged dictionary back to the JSON file.
        try:
            with open(json_path, 'w') as file:
                json.dump(export_dict, file, indent=4)
                file.flush()
        except Exception as e:
            QMessageBox.critical(self, "Export Error", f"Failed to write annotations.json:\n{e}")

    def on_error(self, message):
        QMessageBox.warning(self, "Error", message)

    def on_worker_finished(self):
        if self.progress_bar:
            self.progress_bar.stop_progress()
            self.progress_bar.close()
        self.thread.quit()
        self.thread.wait()
        self.worker.deleteLater()
        self.thread.deleteLater()
        self.worker = None
        self.thread = None
        QApplication.restoreOverrideCursor()
        self.button_box.setEnabled(True)
        self.accept()

    def reject(self):
        if self.thread and self.thread.isRunning():
            self.worker.stop()
            self.button_box.setEnabled(False)
        else:
            super().reject()

    def closeEvent(self, event):
        self.reject()
        super().closeEvent(event)


import warnings

from typing import Any, Dict, List, Optional, Set

from PyQt5.QtCore import Qt, QAbstractTableModel, QModelIndex, QVariant
from PyQt5.QtGui import QFont, QColor, QBrush

from coralnet_toolbox.Rasters import RasterManager

warnings.filterwarnings("ignore", category=DeprecationWarning)


# ----------------------------------------------------------------------------------------------------------------------
# Classes
# ----------------------------------------------------------------------------------------------------------------------


class RasterTableModel(QAbstractTableModel):
    """
    Custom table model for displaying a list of Raster objects.
    """
    # Column indices
    CHECKBOX_COL = 0
    FILENAME_COL = 1
    ANNOTATION_COUNT_COL = 2
    
    # Row colors
    HIGHLIGHTED_COLOR = QColor(173, 216, 230)  # Light blue
    SELECTED_COLOR = QColor(144, 238, 144)     # Light green
    
    def __init__(self, raster_manager: RasterManager, parent=None):
        """
        Initialize the model.
        
        Args:
            raster_manager (RasterManager): Reference to the raster manager
            parent: Parent widget
        """
        super().__init__(parent)
        self.raster_manager = raster_manager
        self.filtered_paths: List[str] = []
        
        self.column_headers = ["\u2713", "Image Name", "Annotations"]
        
        # Column widths
        self.column_widths = [30, -1, 120]  # -1 means stretch
        
        # Connect to manager signals
        self.raster_manager.rasterAdded.connect(self.on_raster_added)
        self.raster_manager.rasterRemoved.connect(self.on_raster_removed)
        self.raster_manager.rasterUpdated.connect(self.on_raster_updated)
        
    def rowCount(self, parent=None) -> int:
        """Return the number of rows in the model."""
        return len(self.filtered_paths)
        
    def columnCount(self, parent=None) -> int:
        """Return the number of columns in the model."""
        return len(self.column_headers)
        
    def headerData(self, section: int, orientation: Qt.Orientation, role: int = Qt.DisplayRole) -> Any:
        """Return header data for the model."""
        if role == Qt.DisplayRole and orientation == Qt.Horizontal:
            return self.column_headers[section]
        return None
        
    def data(self, index: QModelIndex, role: int = Qt.DisplayRole) -> Any:
        """Return data for the given index and role."""
        if not index.isValid() or index.row() >= len(self.filtered_paths):
            return None
            
        # Get the raster for this row
        path = self.filtered_paths[index.row()]
        raster = self.raster_manager.get_raster(path)
        
        if not raster:
            return None
            
        # Make sure display name is set
        raster.set_display_name(max_length=25)
        
        if role == Qt.DisplayRole:
            if index.column() == self.CHECKBOX_COL:
                return "\u2713" if raster.checkbox_state else ""
            elif index.column() == self.FILENAME_COL:
                return raster.display_name
            elif index.column() == self.ANNOTATION_COUNT_COL:
                return str(raster.annotation_count)
                
        elif role == Qt.TextAlignmentRole:
            return Qt.AlignCenter
        
        elif role == Qt.FontRole:
            # Bold the selected raster's text
            if raster.is_selected:
                font = QFont()
                font.setBold(True)
                return font
                
        elif role == Qt.BackgroundRole:
            # Set background color based on selection/highlight state
            if raster.is_selected:
                return QBrush(self.SELECTED_COLOR)
            elif raster.is_highlighted:
                return QBrush(self.HIGHLIGHTED_COLOR)
                
        elif role == Qt.ToolTipRole:
            if index.column() == self.FILENAME_COL:
                dimensions = raster.metadata.get('dimensions', f"{raster.width}x{raster.height}")
                
                tooltip_parts = [
                    f"<b>Path:</b> {path}",
                    f"<b>Dimensions:</b> {dimensions}",
                    f"<b>Annotations:</b> {'Yes' if raster.has_annotations else 'No'}",
                    f"<b>Predictions:</b> {'Yes' if raster.has_predictions else 'No'}"
                ]

                if raster.has_work_areas():
                    tooltip_parts.append(f"<b>Work Areas:</b> {raster.count_work_items()}")

                return "<br>".join(tooltip_parts)

            elif index.column() == self.ANNOTATION_COUNT_COL and raster.annotation_count > 0:
                tooltip_text = f"<b>Total annotations:</b> {raster.annotation_count}"
                
                # Add annotation counts per label using a for loop
                if hasattr(raster, 'label_counts') and raster.label_counts:
                    label_items = []
                    for label, count in raster.label_counts.items():
                        label_items.append(f"<li>{label}: {count}</li>")
                    label_counts_text = "".join(label_items)
                    tooltip_text += f"<br><br><b>Annotations by label:</b><ul>{label_counts_text}</ul>"
                
                # Add annotation counts per type using a for loop
                if hasattr(raster, 'annotation_types') and raster.annotation_types:
                    type_items = []
                    for type_name, count in raster.annotation_types.items():
                        type_items.append(f"<li>{type_name}: {count}</li>")
                    type_counts_text = "".join(type_items)
                    tooltip_text += f"<br><b>Annotations by type:</b><ul>{type_counts_text}</ul>"
                
                return tooltip_text
                
        return None
        
    def flags(self, index: QModelIndex) -> Qt.ItemFlags:
        """Return flags for the given index."""
        if not index.isValid():
            return Qt.NoItemFlags
            
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable
    
    def add_path(self, path: str):
        """
        Efficiently add a single path by signaling a row insertion.
        
        Args:
            path (str): The image path to add.
        """
        if path in self.raster_manager.image_paths and path not in self.filtered_paths:
            # The position for the new row is at the end of the current list
            row_position = len(self.filtered_paths)
            
            # Signal that we are about to insert one row at this position
            self.beginInsertRows(QModelIndex(), row_position, row_position)
            
            # Add the data
            self.filtered_paths.append(path)
            
            # Signal that the insertion is complete
            self.endInsertRows()
            
    def highlight_path(self, path: str, highlighted: bool = True):
        """
        Set the highlight state for a specific path
        
        Args:
            path (str): Image path to highlight/unhighlight
            highlighted (bool): Whether to highlight (True) or unhighlight (False)
        """
        raster = self.raster_manager.get_raster(path)
        if raster:
            # Only update if state is changing
            if raster.is_highlighted != highlighted:
                raster.set_highlighted(highlighted)
                
                # Update the view
                row = self.get_row_for_path(path)
                if row >= 0:
                    self.dataChanged.emit(
                        self.index(row, 0),
                        self.index(row, self.columnCount() - 1)
                    )
                    
    def clear_highlights(self):
        """Clear all highlighted paths"""
        # Find all highlighted rasters
        highlighted_paths = []
        for path in self.filtered_paths:
            raster = self.raster_manager.get_raster(path)
            if raster and raster.is_highlighted:
                highlighted_paths.append(path)
        
        # Unhighlight all paths
        for path in highlighted_paths:
            self.highlight_path(path, False)
            
    def set_highlighted_paths(self, paths: List[str]):
        """
        Set the highlighted state for a list of paths, clearing all others.
        
        Args:
            paths (List[str]): List of image paths to highlight
        """
        # First get all currently highlighted paths
        current_highlighted = self.get_highlighted_paths()
        
        # Unhighlight those that shouldn't be highlighted
        for path in current_highlighted:
            if path not in paths:
                self.highlight_path(path, False)
                
        # Highlight those that should be highlighted
        for path in paths:
            if path in self.filtered_paths:  # Only highlight visible paths
                self.highlight_path(path, True)

    def clear_highlights_except(self, paths_to_keep: List[str]):
        """
        Clear highlights except for the given paths.
        
        Args:
            paths_to_keep (List[str]): Paths to keep highlighted
        """
        for path in self.get_highlighted_paths():
            if path not in paths_to_keep:
                self.highlight_path(path, False)

    def get_highlighted_paths(self) -> List[str]:
        """
        Get a list of all highlighted paths
        
        Returns:
            List[str]: List of highlighted image paths
        """
        highlighted_paths = []
        for path in self.filtered_paths:
            raster = self.raster_manager.get_raster(path)
            if raster and raster.is_highlighted:
                highlighted_paths.append(path)
        return highlighted_paths
        
    def set_filtered_paths(self, paths: List[str]):
        """
        Set the filtered paths to display in the model.
        
        Args:
            paths (List[str]): List of image paths to display
        """
        self.beginResetModel()
        self.filtered_paths = [p for p in paths if p in self.raster_manager.image_paths]
        self.endResetModel()
        
    def get_path_at_row(self, row: int) -> Optional[str]:
        """
        Get the image path at the given row.
        
        Args:
            row (int): Row index
            
        Returns:
            str or None: Image path or None if invalid
        """
        if 0 <= row < len(self.filtered_paths):
            return self.filtered_paths[row]
        return None
        
    def get_row_for_path(self, path: str) -> int:
        """
        Get the row index for the given image path.
        
        Args:
            path (str): Image path
            
        Returns:
            int: Row index, or -1 if not found
        """
        try:
            return self.filtered_paths.index(path)
        except ValueError:
            return -1
            
    def set_selected_path(self, path: str):
        """
        Set the selected path and update the model.
        
        Args:
            path (str): Image path to select
        """
        # Get currently selected paths first
        selected_paths = []
        for p in self.filtered_paths:
            raster = self.raster_manager.get_raster(p)
            if raster and raster.is_selected:
                selected_paths.append(p)
        
        # Clear any previous selection
        for p in selected_paths:
            raster = self.raster_manager.get_raster(p)
            if raster:
                raster.set_selected(False)
                row = self.get_row_for_path(p)
                if row >= 0:
                    self.dataChanged.emit(
                        self.index(row, 0),
                        self.index(row, self.columnCount() - 1)
                    )
        
        # Set new selection
        raster = self.raster_manager.get_raster(path)
        if raster:
            # Only update if state is changing
            if not raster.is_selected:
                raster.set_selected(True)
                row = self.get_row_for_path(path)
                if row >= 0:
                    self.dataChanged.emit(
                        self.index(row, 0),
                        self.index(row, self.columnCount() - 1)
                    )
    
    def get_selected_path(self) -> Optional[str]:
        """
        Get the currently selected path.
        
        Returns:
            str or None: Selected path or None if no selection
        """
        for path in self.filtered_paths:
            raster = self.raster_manager.get_raster(path)
            if raster and raster.is_selected:
                return path
        return None
                
    def update_raster_data(self, path: str):
        """
        Update display for a specific raster.
        
        Args:
            path (str): Image path to update
        """
        row = self.get_row_for_path(path)
        if row >= 0:
            self.dataChanged.emit(
                self.index(row, 0),
                self.index(row, self.columnCount() - 1)
            )
            
    def on_raster_added(self, path: str):
        """Handler for when a raster is added to the manager."""
        # If we want to automatically add new rasters to filtered list:
        # if path not in self.filtered_paths:
        #     self.beginInsertRows(QModelIndex(), len(self.filtered_paths), len(self.filtered_paths))
        #     self.filtered_paths.append(path)
        #     self.endInsertRows()
        pass
        
    def on_raster_removed(self, path: str):
        """Handler for when a raster is removed from the manager."""
        row = self.get_row_for_path(path)
        if row >= 0:
            self.beginRemoveRows(QModelIndex(), row, row)
            self.filtered_paths.remove(path)
            self.endRemoveRows()
            
    def on_raster_updated(self, path: str):
        """Handler for when a raster is updated in the manager."""
        self.update_raster_data(path)
    
    # New methods to better sync with Qt's selection model
    def sync_with_selection_model(self, selected_indexes, deselected_indexes):
        """
        Synchronize with Qt's selection model changes.
        
        Args:
            selected_indexes: Indexes that were selected
            deselected_indexes: Indexes that were deselected
        """
        # Handle deselections
        for index in deselected_indexes:
            if index.isValid():
                path = self.get_path_at_row(index.row())
                if path:
                    raster = self.raster_manager.get_raster(path)
                    if raster and raster.is_selected:
                        raster.set_selected(False)
                        self.update_raster_data(path)
        
        # Handle selections
        for index in selected_indexes:
            if index.isValid():
                path = self.get_path_at_row(index.row())
                if path:
                    raster = self.raster_manager.get_raster(path)
                    if raster and not raster.is_selected:
                        raster.set_selected(True)
                        self.update_raster_data(path)